import shutil

import pytest
import sympy
from sympy import Symbol as symbol

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import (
    Assignment,
    CompartmentalSystem,
    EstimationStep,
    EstimationSteps,
    Model,
    ModelSyntaxError,
    NormalDistribution,
    Parameter,
    Parameters,
    Statements,
)
from pharmpy.model.external.nonmem import convert_model
from pharmpy.model.external.nonmem.nmtran_parser import NMTranParser
from pharmpy.model.external.nonmem.records.factory import create_record
from pharmpy.modeling import (
    add_iiv,
    add_population_parameter,
    create_basic_pk_model,
    create_joint_distribution,
    read_model,
    read_model_from_string,
    remove_iiv,
    set_direct_effect,
    set_estimation_step,
    set_initial_condition,
    set_initial_estimates,
    set_transit_compartments,
    set_zero_order_elimination,
    set_zero_order_input,
)
from pharmpy.tools import read_modelfit_results


def _ensure_trailing_newline(buf):
    # FIXME: This should not be necessary
    return buf if buf[-1] == '\n' else buf + '\n'


def S(x):
    return symbol(x)


def test_source(pheno):
    assert pheno.model_code.startswith('$PROBLEM PHENOBARB')


def test_convert_nonmem_model(testdata):
    model = create_basic_pk_model('iv', testdata / 'nonmem/pheno_pd.csv')
    model = set_direct_effect(model, 'linear')
    model = convert_model(model)
    assert model.dependent_variables == {S('Y'): 1, S('Y_2'): 2}
    assert model.observation_transformation == {S('Y'): S('Y'), S('Y_2'): S('Y_2')}


def test_parse_nonmem_model(testdata):
    model = read_model(testdata / 'nonmem/pheno_pd.mod')
    assert model.dependent_variables == {S('Y'): 1}
    assert model.observation_transformation == {S('Y'): S('Y')}

    model = read_model(testdata / 'nonmem/models/pheno_dvid.mod')
    assert model.dependent_variables == {S('Y_1'): 1, S('Y_2'): 2}
    assert model.observation_transformation == {S('Y_1'): S('Y_1'), S('Y_2'): S('Y_2')}


def test_update_inits(load_model_for_test, pheno_path):
    from pharmpy.modeling import update_inits

    model = load_model_for_test(pheno_path)
    res = read_modelfit_results(pheno_path)
    model = update_inits(model, res.parameter_estimates)


def test_empty_ext_file(testdata):
    # Assert existing but empty ext-file does not give modelfit_results
    res = read_modelfit_results(
        testdata / 'nonmem' / 'modelfit_results' / 'onePROB' / 'noESTwithSIM' / 'onlysim.mod'
    )
    assert res is None


def test_detection():
    Model.parse_model_from_string("$PROBLEM this\n$PRED\n")
    Model.parse_model_from_string("   \t$PROBLEM skld fjl\n$PRED\n")
    Model.parse_model_from_string(" $PRO l907\n$PRED\n")


def test_validate():
    with pytest.raises(ModelSyntaxError):
        Model.parse_model_from_string("$PROBLEM this\n$SIZES LIM1=3000\n$PRED\n")


def test_parameters(pheno):
    params = pheno.parameters
    assert len(params) == 6
    assert pheno.parameters['PTVCL'] == Parameter('PTVCL', 0.00469307, lower=0, upper=float("inf"))
    assert pheno.parameters['PTVV'] == Parameter('PTVV', 1.00916, lower=0, upper=float("inf"))
    assert pheno.parameters['THETA_3'] == Parameter('THETA_3', 0.1, lower=-0.99, upper=float("inf"))
    assert pheno.parameters['IVCL'] == Parameter('IVCL', 0.0309626, lower=0, upper=float("inf"))
    assert pheno.parameters['IVV'] == Parameter('IVV', 0.031128, lower=0, upper=float("inf"))
    assert pheno.parameters['SIGMA_1_1'] == Parameter(
        'SIGMA_1_1', 0.013241, lower=0, upper=float("inf")
    )


def test_set_parameters(pheno_path, load_model_for_test):
    pheno = load_model_for_test(pheno_path)
    res = read_modelfit_results(pheno_path)
    params = {
        'PTVCL': 0.75,
        'PTVV': 0.5,
        'THETA_3': 0.25,
        'IVCL': 0.1,
        'IVV': 0.2,
        'SIGMA_1_1': 0.3,
    }
    model = set_initial_estimates(pheno, params)
    assert model.parameters['PTVCL'] == Parameter('PTVCL', 0.75, lower=0, upper=float("inf"))
    assert model.parameters['PTVV'] == Parameter('PTVV', 0.5, lower=0, upper=float("inf"))
    assert model.parameters['THETA_3'] == Parameter(
        'THETA_3', 0.25, lower=-0.99, upper=float("inf")
    )
    assert model.parameters['IVCL'] == Parameter('IVCL', 0.1, lower=0, upper=float("inf"))
    assert model.parameters['IVV'] == Parameter('IVV', 0.2, lower=0, upper=float("inf"))
    assert model.parameters['SIGMA_1_1'] == Parameter('SIGMA_1_1', 0.3, lower=0, upper=float("inf"))
    model = model.update_source()
    thetas = model.internals.control_stream.get_records('THETA')
    assert str(thetas[0]) == '$THETA (0,0.75) ; PTVCL\n'
    assert str(thetas[1]) == '$THETA (0,0.5) ; PTVV\n'
    assert str(thetas[2]) == '$THETA (-0.99,0.25)\n'
    omegas = model.internals.control_stream.get_records('OMEGA')
    assert str(omegas[0]) == '$OMEGA DIAGONAL(2)\n 0.1  ;       IVCL\n 0.2  ;        IVV\n\n'
    sigmas = model.internals.control_stream.get_records('SIGMA')
    assert str(sigmas[0]) == '$SIGMA 0.3\n'

    model = set_initial_estimates(pheno, {'PTVCL': 18})
    assert model.parameters['PTVCL'] == Parameter('PTVCL', 18, lower=0, upper=float("inf"))
    assert model.parameters['PTVV'] == Parameter('PTVV', 1.00916, lower=0, upper=float("inf"))

    model = create_joint_distribution(pheno, individual_estimates=res.individual_estimates)
    with pytest.raises(UserWarning, match='Adjusting initial'):
        set_initial_estimates(model, {'IVV': 0.000001})


def test_adjust_iovs(load_model_for_test, testdata):
    model = load_model_for_test(
        testdata / 'nonmem' / 'modelfit_results' / 'onePROB' / 'multEST' / 'noSIM' / 'withBayes.mod'
    )
    model.parameters
    rvs = model.random_variables

    assert rvs[0].level == 'IIV'
    assert rvs[3].level == 'IOV'
    assert rvs[4].level == 'IOV'
    assert rvs[6].level == 'IOV'

    model = load_model_for_test(testdata / 'nonmem' / 'qa' / 'iov.mod')
    dists = model.random_variables
    assert dists[0].level == 'IIV'
    assert dists[1].level == 'IOV'
    assert dists[2].level == 'IOV'


@pytest.mark.parametrize(
    'param_new,init_expected,buf_new',
    [
        (Parameter.create('COVEFF', 0.2), 0.2, '$THETA  0.2 ; COVEFF'),
        (Parameter.create('THETA', 0.1), 0.1, '$THETA  0.1 ; THETA'),
        (Parameter.create('THETA', 0.1, 0, fix=True), 0.1, '$THETA  (0,0.1) FIX ; THETA'),
        (Parameter.create('RUV_prop', 0.1), 0.1, '$THETA  0.1 ; RUV_prop'),
    ],
)
def test_add_parameters(pheno, param_new, init_expected, buf_new):
    pset = [p for p in pheno.parameters]

    assert len(pset) == 6

    pset.append(param_new)
    model = pheno.replace(parameters=Parameters.create(pset))

    assert len(pset) == 7
    assert model.parameters[param_new.name].init == init_expected

    rec_ref = (
        f'$THETA (0,0.00469307) ; PTVCL\n'
        f'$THETA (0,1.00916) ; PTVV\n'
        f'$THETA (-.99,.1)\n'
        f'{buf_new}\n'
    )

    model = model.update_source()
    rec_mod = ''
    for rec in model.internals.control_stream.get_records('THETA'):
        rec_mod += str(rec)

    assert rec_ref == rec_mod


def test_add_two_parameters(pheno):
    assert len(pheno.parameters) == 6

    model = add_population_parameter(pheno, 'COVEFF1', 0.2)
    model = add_population_parameter(model, 'COVEFF2', 0.1)

    assert len(model.parameters) == 8
    assert model.parameters['COVEFF1'].init == 0.2
    assert model.parameters['COVEFF2'].init == 0.1


@pytest.mark.parametrize(
    'statement_new,param_new,buf_new',
    [
        (Assignment(S('CL'), sympy.Integer(2)), None, 'CL = 2'),
        (
            Assignment(S('Y'), S('THETA(4)') + S('THETA(5)')),
            [Parameter.create('THETA(4)', 0.1), Parameter.create('THETA(5)', 0.1)],
            'Y = THETA(4) + THETA(5)',
        ),
    ],
)
def test_add_statements(pheno, statement_new, buf_new, param_new):
    sset = pheno.statements
    assert len(sset) == 15

    # Insert new statement before ODE system.
    new_sset = []
    for s in sset:
        if isinstance(s, CompartmentalSystem):
            new_sset.append(statement_new)
        new_sset.append(s)

    model = pheno.replace(
        statements=Statements(new_sset), parameters=pheno.parameters + Parameters.create(param_new)
    )
    model = model.update_source()

    assert len(model.statements) == 16

    parser = NMTranParser()
    stream = parser.parse(model.model_code)

    assert str(model.internals.control_stream) == str(stream)

    rec_ref = (
        f'$PK\n'
        f'IF(AMT.GT.0) BTIME=TIME\n'
        f'TAD=TIME-BTIME\n'
        f'TVCL=THETA(1)*WGT\n'
        f'TVV=THETA(2)*WGT\n'
        f'IF(APGR.LT.5) TVV=TVV*(1+THETA(3))\n'
        f'CL=TVCL*EXP(ETA(1))\n'
        f'V=TVV*EXP(ETA(2))\n'
        f'S1=V\n'
        f'{buf_new}\n\n'
    )

    rec_mod = str(model.internals.control_stream.get_records('PK')[0])

    assert rec_ref == rec_mod


@pytest.mark.parametrize(
    'param_new, statement_new, buf_new',
    [
        (Parameter.create('X', 0.1), Assignment(S('Y'), S('X') + S('S1')), 'Y = S1 + THETA(4)'),
    ],
)
def test_add_parameters_and_statements(pheno, param_new, statement_new, buf_new):
    model = add_population_parameter(pheno, param_new.name, param_new.init)

    sset = model.statements

    # Insert new statement before ODE system.
    new_sset = []
    for s in sset:
        if isinstance(s, CompartmentalSystem):
            new_sset.append(statement_new)
        new_sset.append(s)

    model = model.replace(statements=Statements(new_sset))
    model = model.update_source()

    rec = (
        f'$PK\n'
        f'IF(AMT.GT.0) BTIME=TIME\n'
        f'TAD=TIME-BTIME\n'
        f'TVCL=THETA(1)*WGT\n'
        f'TVV=THETA(2)*WGT\n'
        f'IF(APGR.LT.5) TVV=TVV*(1+THETA(3))\n'
        f'CL=TVCL*EXP(ETA(1))\n'
        f'V=TVV*EXP(ETA(2))\n'
        f'S1=V\n'
        f'{buf_new}\n\n'
    )

    assert str(model.internals.control_stream.get_pred_pk_record()) == rec


@pytest.mark.parametrize('rv_new, buf_new', [(Parameter('omega', 0.1), '$OMEGA  0.1')])
def test_add_random_variables(pheno, rv_new, buf_new):
    rvs = pheno.random_variables

    eta = NormalDistribution.create('eta_new', 'iiv', 0, S(rv_new.name))

    model = add_population_parameter(pheno, rv_new.name, rv_new.init)
    model = model.replace(random_variables=rvs + eta)

    model = model.update_source()

    rec_ref = (
        f'$OMEGA DIAGONAL(2)\n'
        f' 0.0309626  ;       IVCL\n'
        f' 0.031128  ;        IVV\n\n'
        f'{buf_new} ; omega\n'
    )

    rec_mod = ''
    for rec in model.internals.control_stream.get_records('OMEGA'):
        rec_mod += str(rec)

    assert rec_mod == rec_ref

    rv = model.random_variables['eta_new']

    assert rv.mean == 0
    assert rv.variance.name == 'omega'


def test_add_random_variables_and_statements(pheno):
    rvs = pheno.random_variables

    eta = NormalDistribution.create('ETA_NEW', 'iiv', 0, S('omega'))
    rvs = rvs + eta
    model = add_population_parameter(pheno, 'omega', 0.1)

    eps = NormalDistribution.create('EPS_NEW', 'ruv', 0, S('sigma'))
    rvs = rvs + eps
    model = add_population_parameter(model, 'sigma', 0.1)

    sset = model.statements
    statement_new = Assignment(S('X'), 1 + S(eps.names[0]) + S(eta.names[0]))
    model = model.replace(
        statements=sset.before_odes + statement_new + sset.ode_system + sset.after_odes,
        random_variables=rvs,
    )

    model = model.update_source()
    print(model.internals.control_stream.get_pred_pk_record())
    assert str(model.internals.control_stream.get_pred_pk_record()).endswith(
        'X = ETA_NEW + EPS(2) + 1\n\n'
    )
    assert '$ABBR REPLACE ETA_NEW=ETA(1)'


def test_minimal(load_model_for_test, datadir):
    path = datadir / 'minimal.mod'
    model = load_model_for_test(path)
    assert len(model.statements) == 1
    assert model.statements[0].expression == symbol('THETA_1') + symbol('ETA_1') + symbol('EPS_1')


def test_initial_individual_estimates(load_model_for_test, datadir):
    path = datadir / 'minimal.mod'
    model = load_model_for_test(path)
    assert model.initial_individual_estimates is None

    path = datadir / 'pheno_etas.mod'
    model = load_model_for_test(path)
    inits = model.initial_individual_estimates
    assert len(inits) == 59
    assert len(inits.columns) == 2
    assert inits['ETA_1'][2] == -0.166321


def test_deterministic_theta_comments(pheno):
    no_option = 0
    for theta_record in pheno.internals.control_stream.get_records('THETA'):
        no_option += len(list(theta_record.root.subtrees('option')))

    assert no_option == 0


def test_remove_eta(pheno):
    model = remove_iiv(pheno, 'ETA_1')
    assert model.model_code.split('\n')[13] == 'V = TVV*EXP(ETA_2)'
    assert '$ABBR REPLACE ETA_2=ETA(1)'


def test_symbol_names_in_comment(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    assert model.statements[2].expression == S('PTVCL') * S('WGT')

    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
TV = THETA(1)
Y = TV + ETA(1) + ERR(1)

$THETA 0.1  ; TV
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    with pytest.warns(UserWarning):
        model = Model.parse_model_from_string(code)
        assert model.parameters.names == ['THETA_1', 'OMEGA_1_1', 'SIGMA_1_1']


def test_symbol_names_in_abbr(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_abbr.mod')
    pset, rvs = model.parameters, model.random_variables

    assert 'THETA_CL' in pset.names
    assert 'ETA_CL' in rvs.etas.names


def test_clashing_parameter_names(load_model_for_test, datadir):
    with pytest.warns(UserWarning):
        model = load_model_for_test(datadir / 'pheno_clashing_symbols.mod')
    assert model.parameters.names == ['THETA_1', 'TVV', 'IVCL', 'OMEGA_2_2', 'SIGMA_1_1']

    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$THETA 0.1  ; TV
$OMEGA 0.01 ; TV
$SIGMA 1 ; TV
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    with pytest.warns(UserWarning):
        model = Model.parse_model_from_string(code)
        assert model.parameters.names == ['TV', 'OMEGA_1_1', 'SIGMA_1_1']

    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + THETA(2)

$THETA 0.1  ; TV
$THETA 0.1  ; TV
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    with pytest.warns(UserWarning):
        model = Model.parse_model_from_string(code)
        assert model.parameters.names == ['TV', 'THETA_2']


def test_abbr_write(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    model = add_iiv(model, 'S1', 'add')

    assert 'ETA_S1' in model.model_code
    assert 'ETA_S1' in model.random_variables.names
    assert S('ETA_S1') in model.statements.free_symbols


def test_abbr_read_write(load_model_for_test, pheno_path):
    model_write = load_model_for_test(pheno_path)
    model_write = add_iiv(model_write, 'S1', 'add')
    model_read = Model.parse_model_from_string(model_write.model_code)
    assert model_read.model_code == model_write.model_code
    assert model_read.statements == model_write.statements
    assert not (
        set(model_read.random_variables.names) - set(model_write.random_variables.names)
    )  # Different order due to renaming in read


def test_dv_symbol(pheno):
    assert list(pheno.dependent_variables.keys())[0].name == 'Y'


def test_insert_unknown_record(pheno):
    rec = create_record('$TRIREME one')
    newcs = pheno.internals.control_stream.insert_record(rec)
    model = pheno.replace(internals=pheno.internals.replace(control_stream=newcs))
    assert model.model_code.split('\n')[-1] == '$TRIREME one'

    rec = create_record('\n$OA two')
    newcs = model.internals.control_stream.insert_record(rec)
    model = model.replace(internals=model.internals.replace(control_stream=newcs))
    assert model.model_code.split('\n')[-1] == '$OA two'


def test_parse_illegal_record_name():
    parser = NMTranParser()
    with pytest.raises(ModelSyntaxError):
        parser.parse('$PROBLEM MYPROB\n$1REC\n')


def test_frem_model():
    code = """$PROBLEM    MOXONIDINE PK,FINAL ESTIMATES,simulated data
$INPUT      ID VISI DROP DGRP DOSE DROP DROP DROP DROP NEUY SCR AGE
            SEX DROP WT DROP ACE DIG DIU DROP TAD TIME DROP CRCL AMT
            SS II DROP DROP DROP DV DROP DROP MDV FREMTYPE
$DATA      ../frem_dataset.dta IGNORE=@
$SUBROUTINE ADVAN2 TRANS1
$OMEGA  BLOCK(1)
 1.47E-06  ;     IOV_CL
$OMEGA  BLOCK(1) SAME
$OMEGA  BLOCK(1)
 5.06E-05  ;     IOV_KA
$OMEGA  BLOCK(1) SAME
$OMEGA  BLOCK(5)
 0.416046
 0.389578 0.578819  ;   IIV_CL_V
 0.00327757 0.00386591 0.258203  ;     IIV_KA
 -0.0809376 -0.130036 0.0777389 1  ;    BSV_AGE
 0.0880473 0.0821262 -0.0578989 -0.11524 1  ;    BSV_SEX
$PK
   VIS3               = 0
   IF(VISI.EQ.3) VIS3 = 1
   VIS8               = 0
   IF(VISI.EQ.8) VIS8 = 1
   KPCL  = VIS3*ETA(1)+VIS8*ETA(2)
   KPKA  = VIS3*ETA(3)+VIS8*ETA(4)
   TVCL  = THETA(1)
   TVV   = THETA(2)
   CL    = TVCL*EXP(ETA(5)+KPCL)
   V     = TVV*EXP(ETA(6))
   KA    = THETA(3)*EXP(ETA(7)+KPKA)
   ALAG1 = THETA(4)
   K     = CL/V
   S2    = V
    SDC8 = 7.82226906804
    SDC9 = 0.404756978659

$ERROR
     IPRED = LOG(.025)
     WA     = 1
     W      = WA
     IF(F.GT.0) IPRED = LOG(F)
     IRES  = IPRED-DV
     IWRES = IRES/W
     Y     = IPRED+ERR(1)*W

;;;FREM CODE BEGIN COMPACT
;;;DO NOT MODIFY
    IF (FREMTYPE.EQ.100) THEN
;      AGE  7.82226906804
       Y = THETA(5) + ETA(8)*SDC8 + EPS(2)
       IPRED = THETA(5) + ETA(8)*SDC8
    END IF
    IF (FREMTYPE.EQ.200) THEN
;      SEX  0.404756978659
       Y = THETA(6) + ETA(9)*SDC9 + EPS(2)
       IPRED = THETA(6) + ETA(9)*SDC9
    END IF
;;;FREM CODE END COMPACT
$THETA  (0,32.901) ; POP_TVCL
$THETA  (0,115.36) ; POP_TVV
$THETA  (0,1.45697) ; POP_TVKA
$THETA  (0,0.0818029) ; POP_LAG
$THETA  65.1756756757 FIX ; TV_AGE
 1.2027027027 FIX ; TV_SEX
$SIGMA  0.112373
$SIGMA  0.0000001  FIX  ;     EPSCOV
$ESTIMATION METHOD=1 MAXEVAL=9999 NONINFETA=1 MCETA=1
"""
    model = Model.parse_model_from_string(code)
    rvs = model.random_variables
    assert len(rvs.names) == 11


@pytest.mark.parametrize(
    'model_path, transformation',
    [
        ('nonmem/pheno.mod', set_zero_order_elimination),
    ],
)
def test_des(load_model_for_test, testdata, model_path, transformation):
    model_ref = load_model_for_test(testdata / model_path)
    model_ref = transformation(model_ref)

    model_des = Model.parse_model_from_string(model_ref.model_code)

    assert model_ref.statements.ode_system == model_des.statements.ode_system


def test_cmt_update(load_model_for_test, testdata, tmp_path):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'mox1.mod')
    old_cmt = model.dataset["CMT"]

    # CMT is dropped --> No update
    model_transits = set_transit_compartments(model, 2)
    updated_cmt = model_transits.dataset["CMT"]
    assert all(old_cmt == updated_cmt)

    # CMT undropped. Require generation of new column
    from pharmpy.modeling import undrop_columns

    model = undrop_columns(model, 'CMT')
    model_transits = set_transit_compartments(model, 2)
    updated_cmt = model_transits.dataset["CMT"]
    assert set(updated_cmt.unique()) == {0, 1}

    with chdir(tmp_path):
        dataset = model.dataset.copy()
        dataset["CMT"] = 2
        dataset.loc[(dataset["ID"] <= 573) & (dataset["AMT"] != 0), "CMT"] = 1
        dataset.to_csv(str(tmp_path / 'temp_dataset.csv'), index=False)
        model_string = model.model_code
        model_string = model_string.replace("DUMMYPATH", str(tmp_path / 'temp_dataset.csv'))
        model_string = model_string.replace("CMT=DROP", "CMT")

        model = read_model_from_string(model_string)
        assert set(model.dataset.CMT.unique()) == {1, 2}

        model = set_transit_compartments(model, 2)
        assert (set(model.dataset.CMT.unique())) == {1, 4}

        model = set_transit_compartments(model, 1)
        assert (set(model.dataset.CMT.unique())) == {1, 3}


@pytest.mark.parametrize(
    'estcode,est_steps',
    [
        ('$ESTIMATION METH=COND INTERACTION', [EstimationStep.create('foce', interaction=True)]),
        ('$ESTIMATION INTER METH=COND', [EstimationStep.create('foce', interaction=True)]),
        ('$ESTM METH=1 INTERACTION', [EstimationStep.create('foce', interaction=True)]),
        ('$ESTIM METH=1', [EstimationStep.create('foce')]),
        ('$ESTIMA METH=0', [EstimationStep.create('fo')]),
        ('$ESTIMA METH=ZERO', [EstimationStep.create('fo')]),
        ('$ESTIMA INTER', [EstimationStep.create('fo', interaction=True)]),
        (
            '$ESTIMA INTER\n$COV',
            [
                EstimationStep.create(
                    'fo', interaction=True, parameter_uncertainty_method='sandwich'
                )
            ],
        ),
        (
            '$ESTIMA METH=COND INTER\n$EST METH=COND',
            [
                EstimationStep.create('foce', interaction=True),
                EstimationStep.create('foce', interaction=False),
            ],
        ),
        ('$ESTIMATION METH=SAEM', [EstimationStep.create('saem')]),
        ('$ESTIMATION METH=1 LAPLACE', [EstimationStep.create('foce', laplace=True)]),
        (
            '$ESTIMATION METH=0 MAXEVAL=0',
            [EstimationStep.create('fo', evaluation=True)],
        ),
        (
            '$ESTIMATION METH=IMP EONLY=1',
            [EstimationStep.create('imp', evaluation=True)],
        ),
        (
            '$ESTIMATION METH=COND MAXEVAL=9999',
            [EstimationStep.create('foce', maximum_evaluations=9999)],
        ),
        (
            '$ESTIMATION METH=COND ISAMPLE=10 NITER=5 AUTO=1 PRINT=2',
            [EstimationStep.create('foce', isample=10, niter=5, auto=True, keep_every_nth_iter=2)],
        ),
    ],
)
def test_estimation_steps_getter(estcode, est_steps):
    code = '''$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@
$PRED
Y = THETA(1) + ETA(1) + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
'''
    code += estcode
    model = Model.parse_model_from_string(code)
    correct = EstimationSteps.create(steps=est_steps)
    assert model.estimation_steps == correct


def test_estimation_steps_getter_options():
    code = '''$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@
$PRED
Y = THETA(1) + ETA(1) + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 SADDLE_RESET=1
'''
    model = Model.parse_model_from_string(code)
    assert model.estimation_steps[0].method == 'FOCE'
    assert model.estimation_steps[0].tool_options['SADDLE_RESET'] == '1'


@pytest.mark.parametrize(
    'estcode,kwargs,rec_ref',
    [
        ('$EST METH=COND INTER', {'method': 'fo'}, '$ESTIMATION METHOD=ZERO INTER'),
        ('$EST METH=COND INTER', {'interaction': False}, '$ESTIMATION METHOD=COND'),
        (
            '$EST METH=COND INTER',
            {'parameter_uncertainty_method': 'sandwich'},
            '$COVARIANCE UNCONDITIONAL PRINT=E PRECOND=1',
        ),
        (
            '$EST METH=COND INTER',
            {'parameter_uncertainty_method': 'cpg'},
            '$COVARIANCE MATRIX=S UNCONDITIONAL PRINT=E PRECOND=1',
        ),
        (
            '$EST METH=COND INTER',
            {'parameter_uncertainty_method': 'ofim'},
            '$COVARIANCE MATRIX=R UNCONDITIONAL PRINT=E PRECOND=1',
        ),
        (
            '$EST METH=COND INTER MAXEVAL=99999',
            {'method': 'fo'},
            '$ESTIMATION METHOD=ZERO INTER MAXEVAL=99999',
        ),
        (
            '$EST METH=COND INTER POSTHOC',
            {'method': 'fo'},
            '$ESTIMATION METHOD=ZERO INTER POSTHOC',
        ),
        ('$EST METH=COND INTER', {'laplace': True}, '$ESTIMATION METHOD=COND LAPLACE INTER'),
        (
            '$EST METH=COND INTER',
            {'isample': 10, 'niter': 5},
            '$ESTIMATION METHOD=COND INTER ISAMPLE=10 NITER=5',
        ),
        (
            '$EST METH=COND INTER',
            {'auto': True, 'keep_every_nth_iter': 2},
            '$ESTIMATION METHOD=COND INTER AUTO=1 PRINT=2',
        ),
        ('$EST METH=COND INTER', {'auto': False}, '$ESTIMATION METHOD=COND INTER AUTO=0'),
    ],
)
def test_estimation_steps_setter(estcode, kwargs, rec_ref):
    code = '''$PROBLEM base model
$INPUT ID DV TIME
$DATA tests/testdata/nonmem/file.csv IGNORE=@
$PRED
Y = THETA(1) + ETA(1) + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
'''
    code += estcode
    model = Model.parse_model_from_string(code)
    steps = model.estimation_steps
    newstep = steps[0].replace(**kwargs)
    model = model.replace(estimation_steps=newstep + steps[1:])
    model = model.update_source()
    assert model.model_code.split('\n')[-2] == rec_ref


@pytest.mark.parametrize(
    'estcode,kwargs,error_msg',
    [
        (
            '$EST METH=COND MAXEVAL=0',
            {'tool_options': {'MAXEVAL': 999}},
            'MAXEVAL already set as attribute in estimation method object',
        ),
        (
            '$EST METH=COND INTER',
            {'tool_options': {'INTERACTION': None}},
            'INTERACTION already set as attribute in estimation method object',
        ),
    ],
)
def test_set_estimation_steps_option_clash(estcode, kwargs, error_msg):
    code = '''$PROBLEM base model
$INPUT ID DV TIME
$DATA tests/testdata/nonmem/file.csv IGNORE=@
$PRED
Y = THETA(1) + ETA(1) + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
'''
    code += estcode
    model = Model.parse_model_from_string(code)

    steps = model.estimation_steps
    newstep = steps[0].replace(**kwargs)
    newsteps = newstep + steps[1:]
    model = model.replace(estimation_steps=newsteps)

    with pytest.raises(ValueError) as excinfo:
        model.update_source()
    assert error_msg == str(excinfo.value)


def test_add_estimation_step():
    code = '''$PROBLEM base model
$INPUT ID DV TIME
$DATA tests/testdata/nonmem/file.csv IGNORE=@
$PRED
Y = THETA(1) + ETA(1) + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$EST METH=COND INTER
'''
    model = Model.parse_model_from_string(code)
    est_new = EstimationStep.create('IMP', interaction=True, tool_options={'saddle_reset': 1})
    model = model.replace(estimation_steps=model.estimation_steps + est_new)
    model = model.update_source()
    assert model.model_code.split('\n')[-2] == '$ESTIMATION METHOD=IMP INTER SADDLE_RESET=1'
    est_new = EstimationStep.create('SAEM', interaction=True)
    model = model.replace(estimation_steps=est_new + model.estimation_steps)
    model = model.update_source()
    assert model.model_code.split('\n')[-4] == '$ESTIMATION METHOD=SAEM INTER'
    est_new = EstimationStep.create('FO', evaluation=True)
    model = model.replace(estimation_steps=model.estimation_steps + est_new)
    model = model.update_source()
    assert model.model_code.split('\n')[-2] == '$ESTIMATION METHOD=ZERO MAXEVAL=0'
    est_new = EstimationStep.create('IMP', evaluation=True)
    model = model.replace(estimation_steps=model.estimation_steps + est_new)
    model = model.update_source()
    assert model.model_code.split('\n')[-2] == '$ESTIMATION METHOD=IMP EONLY=1'


def test_remove_estimation_step():
    code = '''$PROBLEM base model
$INPUT ID DV TIME
$DATA tests/testdata/nonmem/file.csv IGNORE=@
$PRED
Y = THETA(1) + ETA(1) + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$EST METH=COND INTER
'''
    model = Model.parse_model_from_string(code)
    model = model.replace(estimation_steps=model.estimation_steps[1:])
    assert not model.estimation_steps
    model = model.update_source()
    assert model.model_code.split('\n')[-2] == '$SIGMA 1'


def test_update_source_comments():
    code = """
$PROBLEM    run3.mod PHENOBARB SIMPLE MODEL
$DATA      ../../pheno.dta IGNORE=@
$INPUT      ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK


IF(AMT.GT.0) BTIME=TIME
TAD=TIME-BTIME
      TVCL=THETA(1)*WGT
      TVV=THETA(2)*WGT
IF(APGR.LT.5) TVV=TVV*(1+THETA(3))
      CL=TVCL*EXP(ETA(1))
      V=TVV*EXP(ETA(2))
      S1=V
$ERROR

      W=F
      Y=F+W*EPS(1)

      IPRED=F         ;  individual-specific prediction
      IRES=DV-IPRED   ;  individual-specific residual
      IWRES=IRES/W    ;  individual-specific weighted residual

$THETA  (0,0.00469555) ; CL
$THETA  (0,0.984258) ; V
$THETA  (-.99,0.15892)
$OMEGA  DIAGONAL(2)
 0.0293508  ;       IVCL
 0.027906  ;        IVV
$SIGMA  0.013241
$ESTIMATION METHOD=1 INTERACTION MAXEVALS=9999
$COVARIANCE UNCONDITIONAL
$TABLE      ID TIME AMT WGT APGR IPRED PRED TAD CWRES NPDE NOAPPEND
            NOPRINT ONEHEADER FILE=mytab3
"""
    with pytest.warns(UserWarning):
        model = Model.parse_model_from_string(code)
        model.update_source()


def test_convert_model(testdata):
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + EPS(1)

$THETA 1  ; TH1
$OMEGA 2 ; OM1
$SIGMA 3 ; SI1
$ESTIMATION METHOD=1 INTER
"""
    base = Model.parse_model_from_string(code)
    base.dataset_path = testdata / 'nonmem' / 'file.csv'
    model = convert_model(base)
    model.dataset_path = "file.csv"  # Otherwise we get full path
    correct = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + EPS(1)

$THETA 1  ; TH1
$OMEGA 2 ; OM1
$SIGMA 3 ; SI1
$ESTIMATION METHOD=1 INTER
"""
    assert model.model_code == correct


def test_table_long_ids(testdata):
    code = f"""$PROBLEM base model
    $INPUT ID TIME AMT WGT APGR DV FA1 FA2
    $DATA {testdata / "nonmem" / "pheno.dta"} IGNORE=@

    $PRED
    Y = THETA(1) + ETA(1) + EPS(1)

    $THETA 1  ; TH1
    $OMEGA 2 ; OM1
    $SIGMA 3 ; SI1
    $ESTIMATION METHOD=1 INTER
    """
    model = Model.parse_model_from_string(code)
    dataset_new = model.dataset.copy()
    dataset_new['ID'] = dataset_new['ID'] * 10000
    model = model.replace(dataset=dataset_new)
    model = set_estimation_step(model, 'FO', residuals=['CWRES'])
    assert 'FORMAT=' in model.model_code


def test_convert_model_iv(testdata, tmp_path):
    # FIXME: Move to unit test for amd?
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno_rate.dta', '.')
        start_model = create_basic_pk_model(administration='iv', dataset_path='pheno_rate.dta')
        convert_model(start_model)


def test_parse_derivatives(load_model_for_test, testdata):
    model = load_model_for_test(
        testdata / "nonmem" / "linearize" / "linearize_dir1" / "scm_dir1" / "derivatives.mod"
    )
    assert model.estimation_steps[0].eta_derivatives == ('ETA_1', 'ETA_2')
    assert model.estimation_steps[0].epsilon_derivatives == ('EPS_1',)


def test_no_etas_in_model(pheno):
    pheno = remove_iiv(pheno)
    assert 'DUMMYETA' in pheno.model_code
    assert 'ETA(1)' in pheno.model_code


def test_0_fix_diag_omega():
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + EPS(1)

$THETA 1  ; TH1
$OMEGA 0 FIX ; OM1
$SIGMA 3 ; SI1
$ESTIMATION METHOD=1 INTER
"""
    model = Model.parse_model_from_string(code)
    assert len(model.random_variables.etas) == 1


def test_solver():
    code = """
$PROBLEM
$DATA ../../pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN6 TOL=5
$MODEL COMP=(CENTRAL)
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V
$DES
DADT(1) = -CL/V * A(1)
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469555)
$THETA (0,0.984258)
$THETA (0.15892)
$OMEGA 0.0293508 0.027906
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION MAXEVALS=9999
"""
    model = Model.parse_model_from_string(code)
    assert len(model.estimation_steps) == 1
    step = model.estimation_steps[0]
    assert step.solver == 'DVERK'
    assert step.solver_rtol == 5
    assert step.solver_atol == 1e-12

    code = """
$PROBLEM
$DATA ../../pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN13 TOL=5 ATOL=1.5
$MODEL COMP=(CENTRAL)
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V
$DES
DADT(1) = -CL/V * A(1)
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469555)
$THETA (0,0.984258)
$THETA (0.15892)
$OMEGA 0.0293508 0.027906
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION MAXEVALS=9999
"""
    model = Model.parse_model_from_string(code)
    assert len(model.estimation_steps) == 1
    step = model.estimation_steps[0]
    assert step.solver == 'LSODA'
    assert step.solver_rtol == 5
    assert step.solver_atol == 1.5


def test_if_in_des():
    # Also test parsing of A_0(N) =
    code = """
$PROBLEM
$DATA ../../pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN6 TOL=5
$MODEL COMP=(CENTRAL)
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V
A_0(1) = 2
$DES
IF (CL>0) THEN
DADT(1) = -CL/V * A(1)
ELSE
DADT(1) = -1
END IF
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469555)
$THETA (0,0.984258)
$THETA (0.15892)
$OMEGA 0.0293508 0.027906
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION MAXEVALS=9999
"""
    model = Model.parse_model_from_string(code)
    assert isinstance(model.statements.ode_system.eqs[0].rhs, sympy.Piecewise)
    assert model.statements[3].symbol == sympy.Function("A_CENTRAL")(0)
    assert model.statements[3].expression == sympy.Integer(2)


@pytest.mark.parametrize(
    'estline,likelihood',
    [
        ('$ESTIMATION METHOD=1 LIKELIHOOD INTER', 'LIKELIHOOD'),
        ('$ESTIMATION METHOD=1 -2LL INTER', '-2LL'),
    ],
)
def test_likelihood(estline, likelihood):
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + EPS(1)

$THETA 1  ; TH1
$OMEGA 0 FIX ; OM1
$SIGMA 3 ; SI1
"""
    model = Model.parse_model_from_string(code + estline)
    assert model.value_type == likelihood


def test_f_flag():
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
F_FLAG = 1
Y = THETA(1) + ETA(1) + EPS(1)

$THETA 1  ; TH1
$OMEGA 0 FIX ; OM1
$SIGMA 3 ; SI1
$ESTIMATION METHOD=1 INTER
"""
    model = Model.parse_model_from_string(code)
    assert model.value_type == sympy.Symbol('F_FLAG')


def test_datainfo_model_drop_clash(testdata):
    datapath = testdata / 'nonmem' / 'pheno.dta'
    code = f"""$PROBLEM
$DATA {datapath} IGNORE=@
$INPUT ID TIME AMT WGT=DROP APGR DV
$SUBROUTINE ADVAN1 TRANS2

$PK
CL=THETA(1)

$ERROR
Y=F+F*EPS(1)

$THETA (0,0.00469307) ; TVCL
$SIGMA 0.013241

$ESTIMATION METHOD=1 INTERACTION
"""
    with pytest.warns(
        UserWarning, match='NONMEM .mod and dataset .datainfo disagree on DROP for columns WGT'
    ):
        model = Model.parse_model_from_string(code)

    assert model.datainfo['WGT'].drop


def test_abbr_etas():
    code = '''$PROBLEM
$INPUT ID DV TIME
$DATA file.csv IGNORE=@
$ABBR REPLACE ETA_MY=ETA(1)
$PRED
VAR = ETA_MY
Y = THETA(1) + VAR + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
'''
    model = Model.parse_model_from_string(code)
    assert model.random_variables.etas.names == ['ETA_MY']

    model = add_iiv(model, ['Y'], 'exp', '+', eta_names=['ETA_DUMMY'])
    model = remove_iiv(model, ['ETA_MY'])
    model = add_iiv(model, ['VAR'], 'exp', '+', eta_names=['ETA_MY'])
    assert model.model_code.split('\n')[3] == '$ABBR REPLACE ETA_DUMMY=ETA(1)'
    assert model.model_code.split('\n')[4] == '$ABBR REPLACE ETA_MY=ETA(2)'
    assert not model.model_code.split('\n')[5].startswith('$ABBR')
    assert 'VAR = EXP(ETA_MY)' in model.model_code
    assert 'Y = THETA(1) + VAR + ERR(1) + EXP(ETA_DUMMY)'
    model = remove_iiv(model, ['ETA_DUMMY'])
    assert model.model_code.split('\n')[3] == '$ABBR REPLACE ETA_MY=ETA(1)'
    assert not model.model_code.split('\n')[4].startswith('$ABBR')


def test_abbr_not_replace():
    code = '''$PROBLEM
$INPUT ID DV TIME
$DATA file.csv IGNORE=@
$ABBR PROTECT DERIV2=NO
$PRED
VAR = ETA(1)
Y = THETA(1) + VAR + ERR(1)
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
'''
    model = Model.parse_model_from_string(code)
    model = add_iiv(model, ['Y'], 'exp', '+', eta_names=['ETA_DUMMY'])
    assert model.model_code.split('\n')[3] == '$ABBR PROTECT DERIV2=NO'


def test_parse_dvid(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_dvid.mod')
    assert model.statements[-1] == Assignment.create("Y_2", "EPS_1 * F + EPS_2 + F")
    model = model.update_source()
    assert model.statements[-1] == Assignment.create("Y_2", "EPS_1 * F + EPS_2 + F")


def test_parse_observation_transformation(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_dvid.mod')
    assert model.observation_transformation == {
        sympy.Symbol('Y_1'): sympy.Symbol('Y_1'),
        sympy.Symbol('Y_2'): sympy.Symbol('Y_2'),
    }


def test_validate_eta_names():
    code = '''$PROBLEM
    $INPUT ID DV TIME
    $DATA file.csv IGNORE=@
    $PRED
    Y = THETA(1) + ETA(1) + ERR(1)
    $THETA 0.1
    $OMEGA 0.01
    $SIGMA 1
    '''
    model = Model.parse_model_from_string(code)

    with pytest.raises(ValueError, match='NONMEM does not allow etas named `eta`'):
        add_iiv(model, ['Y'], 'exp', '+', eta_names=['eta'])


def test_ics(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    model = set_initial_condition(model, 'CENTRAL', 23)
    pk = model.internals.control_stream.get_pred_pk_record()
    a = str(pk).split('\n')
    assert a[9] == 'A_0(1) = 23'


def test_zo(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    model = set_zero_order_input(model, "CENTRAL", 10)
    des = model.internals.control_stream.get_records("DES")[0]
    print(des)
    assert str(des) == "$DES\nDADT(1) = -A(1)*CL/V + 10\n"


def test_des_assignments(load_model_for_test, testdata):
    model = load_model_for_test(testdata / "nonmem" / "models" / "pheno_des_assignments.mod")

    stats = model.statements

    assert stats[3] == Assignment.create("KE", "CL/VC")
    assert stats[4] == Assignment.create("EXTRA", "2 * A_CENTRAL(t)")



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Generic, List, Optional, TypeVar, Union

from pharmpy.internals.immutable import Immutable

from .task import Task

if TYPE_CHECKING:
    import networkx as nx
else:
    from pharmpy.deps import networkx as nx

T = TypeVar('T')


class WorkflowBase:
    """Common base class for Workflow and WorkflowBuilder"""

    _g: nx.DiGraph

    @property
    def tasks(self) -> List[Task]:
        """All tasks in workflow"""
        return list(self._g.nodes())

    @property
    def input_tasks(self) -> List[Task]:
        """All input (source) tasks of entire workflow"""
        return [node for node in self._g.nodes if self._g.in_degree(node) == 0]

    @property
    def output_tasks(self) -> List[Task]:
        """All output tasks (sink) tasks of entire workflow"""
        return [node for node in self._g.nodes if self._g.out_degree(node) == 0]

    def get_upstream_tasks(self, task: Task) -> List[Task]:
        """Get all tasks upstream of a certain task

        Parameters
        ----------
        task : Task
            Downstream task

        Returns
        -------
        list
            Upstream tasks
        """
        edges = nx.edge_dfs(self._g, task, orientation='reverse')
        return [node for node, _, _ in edges]

    def get_predecessors(self, task: Task) -> List[Task]:
        """Get all predecessors of task

        Parameters
        ----------
        task : Task
            Downstream task

        Returns
        -------
        list
            Predecessors of task
        """
        return list(self._g.predecessors(task))

    def get_successors(self, task: Task) -> List[Task]:
        """Get all successors of task

        Parameters
        ----------
        task : Task
            Upstream task

        Returns
        -------
        list
            Successors of task
        """
        return list(self._g.successors(task))

    def __len__(self):
        return len(self._g.nodes)


class WorkflowBuilder(WorkflowBase):
    """Builder class for Workflow"""

    def __init__(self, workflow=None, tasks=None, name=None):
        if workflow:
            self._g = workflow._g.copy()
            self.name = workflow.name
        else:
            self._g = nx.DiGraph()
            self.name = name
        if tasks:
            for task in tasks:
                self.add_task(task)

    def add_task(self, task: Task, predecessors: Optional[Union[Task, List[Task]]] = None):
        """Add a task to the workflow

        Predecessors will be connected if given.

        Parameters
        ----------
        task : Task
            Task to add
        predecessors : list or Task
            One or multiple predecessor tasks to connect to the added task
        """
        self._g.add_node(task)
        if predecessors is not None:
            if not isinstance(predecessors, list):
                self._g.add_edge(predecessors, task)
            else:
                for pred in predecessors:
                    self._g.add_edge(pred, task)

    def replace_task(self, task: Task, new_task: Task):
        """Replace a task with a new task

        Parameters
        ----------
        task : Task
            Task to replace
        new_task : Task
            New task
        """
        mapping = {task: new_task}
        nx.relabel_nodes(self._g, mapping, copy=False)

    def insert_workflow(
        self, other: Workflow, predecessors: Optional[Union[Task, List[Task]]] = None
    ):
        """Insert other workflow

        Parameters
        ----------
        other : Workflow
            Workflow to insert
        predecessors : list or Task
            One or multiple predecessor tasks to connect to the inputs
            of the inserted workflow. If None all output tasks will be found
            and used as predecessors.
        """
        if predecessors is None:
            output_tasks = self.output_tasks
        else:
            if isinstance(predecessors, list):
                output_tasks = predecessors
            else:
                output_tasks = [predecessors]
        input_tasks = other.input_tasks
        self._g = nx.compose(self._g, other._g)
        if len(input_tasks) == len(output_tasks):
            for inp, outp in zip(input_tasks, output_tasks):
                self._g.add_edge(outp, inp)
        elif len(input_tasks) == 1:
            for outp in output_tasks:
                self._g.add_edge(outp, input_tasks[0])
        elif len(output_tasks) == 1:
            for inp in input_tasks:
                self._g.add_edge(output_tasks[0], inp)
        else:
            raise ValueError('Having N:M connections between workflows is currently not supported')

    def __add__(self, other: Workflow):
        wb_new = WorkflowBuilder()
        wb_new._g = nx.compose(self._g, other._g)
        wb_new.name = self.name
        return wb_new


class Workflow(WorkflowBase, Generic[T], Immutable):
    """Workflow class

    Representation of a directed acyclic graph with Tasks as nodes and
    the flow of parameters as edges.

    Parameters
    ----------
    tasks : list
        List of tasks for initialization
    """

    def __init__(self, builder=None, graph=None, name: Optional[str] = None):
        if builder is not None:
            self._g = nx.freeze(builder._g.copy())
            self._name = builder.name
        elif graph is not None:
            self._g = graph
            self._name = name

    @classmethod
    def create(cls, builder=None, graph=None, name=None):
        return cls(builder=builder, graph=graph, name=name)

    def replace(self, **kwargs):
        builder = kwargs.get('builder', None)
        if builder is None:
            g = kwargs.get('graph', self._g)
            name = kwargs.get('name', self._name)
        else:
            g = None
            name = None
        return Workflow.create(builder=builder, graph=g, name=name)

    def as_dask_dict(self):
        """Create a dask workflow dictionary

        Returns
        -------
        dict
            Dask workflow dictionary
        """
        ids = {}
        for task in self._g.nodes():
            assert isinstance(task, Task)
            ids[task] = f'{task.name}-{uuid.uuid4()}'

        if len(self.output_tasks) == 1:
            ids[self.output_tasks[0]] = 'results'
        else:
            raise ValueError("Workflow can only have one output task")

        as_dict = {}
        for task in nx.dfs_tree(self._g):
            key = ids[task]
            input_list = list(task.task_input)
            input_list.extend(ids[t] for t in self._g.predecessors(task))
            value = (task.function, *input_list)
            as_dict[key] = value
        return as_dict

    @property
    def name(self):
        """Name of Workflow"""
        return self._name

    def plot_dask(self, filename: str):
        """Save a visualization of workflow to file

        Parameters
        ----------
        filename : str
            Path
        """
        from dask import visualize  # pyright: ignore [reportPrivateImportUsage]

        visualize(self.as_dask_dict(), filename=filename, collapse_outputs=True)

    def __str__(self):
        return '\n'.join([str(task) for task in self._g])

    def __add__(self, other: Workflow):
        g = nx.compose(self._g, other._g)
        wf_new = Workflow(graph=g)
        return wf_new



import pytest
from lark.exceptions import UnexpectedToken

from pharmpy.model import ModelSyntaxError


def test_data_filename_get(parser):
    record = parser.parse('$DATA "pheno.dta"').records[0]
    assert record.filename == 'pheno.dta'

    record = parser.parse('$DATA /home/full/pheno.dta').records[0]
    assert record.filename == '/home/full/pheno.dta'

    record = parser.parse("$DATA 'pheno.dta'").records[0]
    assert str(record.root.subtree('filename')) == "'pheno.dta'"
    assert record.filename == "pheno.dta"

    record = parser.parse(r'$DATA "C:\windowspath\with space in.csv"').records[0]
    assert record.filename == r'C:\windowspath\with space in.csv'

    record = parser.parse('$DATA \n pheno.dta \n; comment\n').records[0]
    assert record.filename == 'pheno.dta'

    record = parser.parse('$DATA ; comment\n ; some comment line\n pheno.dta\n\n').records[0]
    assert record.filename == 'pheno.dta'


def test_data_filename_set(parser):
    record = parser.parse('$DATA DUMMY ; comment').records[0]
    assert record.filename == 'DUMMY'
    assert str(record) == '$DATA DUMMY ; comment'

    # Simple replace
    record = record.set_filename('/new/path/to_file.txt')
    assert record.filename == '/new/path/to_file.txt'
    assert str(record) == '$DATA /new/path/to_file.txt ; comment'

    # Force quoting
    record = record.set_filename('MUST=QUOTE')
    assert record.filename == 'MUST=QUOTE'
    assert str(record) == "$DATA 'MUST=QUOTE' ; comment"

    # More complex example
    text = '$DATA ; comment\n ; some comment line\n pheno.dta\n\n'
    record = parser.parse(text).records[0]
    assert record.filename == 'pheno.dta'
    assert str(record) == text

    # More complex replace
    record = record.set_filename("'IGNORE'")
    assert record.filename == "'IGNORE'"
    assert str(record) == text.replace('pheno.dta', '"\'IGNORE\'"')

    # *
    record = parser.parse('$DATA DUMMY ; comment').records[0]
    record = record.set_filename(None)
    assert str(record) == '$DATA * ; comment'


def test_option_record(parser):
    record = parser.parse('$DATA pheno.dta NOWIDE').records[0]
    assert record.option_pairs == {'NOWIDE': None}


def test_ignore_character(parser):
    record = parser.parse('$DATA pheno.dta').records[0]
    assert record.ignore_character is None
    record = record.set_ignore_character('I')
    assert record.ignore_character == 'I'

    record = parser.parse('$DATA pheno.dta IGNORE=@').records[0]
    assert record.filename == 'pheno.dta'
    assert record.ignore_character == '@'
    record = record.set_ignore_character('K')
    assert record.ignore_character == 'K'

    record = parser.parse('$DATA pheno.dta IGNORE="I"').records[0]
    assert record.ignore_character == 'I'

    record = parser.parse('$DATA pheno.dta IGNORE=\'"\'').records[0]
    assert record.ignore_character == '"'

    record = parser.parse('$DATA pheno.dta IGNORE=K IGNORE=(ID.EQ.2)').records[0]
    assert record.ignore_character == 'K'

    record = parser.parse('$DATA pheno.dta IGNORE=(DV==3) IGNORE=C').records[0]
    assert record.ignore_character == 'C'
    record = record.set_ignore_character('@')
    assert record.ignore_character == '@'
    assert str(record.ignore[0]) == 'DV==3'

    record = parser.parse('$DATA pheno.dta IGNORE=,').records[0]
    assert record.ignore_character == ','

    record = parser.parse('$DATA pheno.dta IGNORE="').records[0]
    assert record.ignore_character == '"'
    record = record.set_ignore_character('"')
    assert record.ignore_character == '"'
    assert str(record) == '$DATA pheno.dta IGNORE="'

    with pytest.raises(UnexpectedToken):
        record = parser.parse('$DATA pheno.dta IGNORE=""').records[0]

    record = parser.parse('$DATA pheno.dta IGNORE=c IGNORE=@').records[0]
    with pytest.raises(ModelSyntaxError):
        record.ignore_character


def test_ignore_character_from_header(parser):
    record = parser.parse('$DATA pheno.dta').records[0]
    assert record.ignore_character is None
    record = record.set_ignore_character_from_header("ID")
    assert record.ignore_character == '@'
    record = record.set_ignore_character_from_header("_ID")
    assert record.ignore_character == '_'


def test_null_value(parser):
    record = parser.parse('$DATA pheno.dta NULL=1').records[0]
    assert record.null_value == 1
    record = parser.parse('$DATA pheno.dta NULL=+').records[0]
    assert record.null_value == 0


def test_ignore_accept(parser):
    record = parser.parse('$DATA pheno.dta IGNORE=(DV.EQ.1)').records[0]
    assert str(record.ignore[0]) == 'DV.EQ.1'
    assert record.accept == []
    record = record.remove_ignore()
    assert record.ignore == []
    assert record.accept == []

    record = parser.parse('$DATA pheno.dta ACCEPT=(DV.EQ.1,    MDV.NEN.23)').records[0]
    assert str(record.accept[0]) == 'DV.EQ.1'
    assert str(record.accept[1]) == 'MDV.NEN.23'
    assert record.ignore == []
    record = record.remove_accept()
    assert record.ignore == []
    assert record.accept == []
    record = parser.parse('$DATA pheno.dta IGNORE=(WGT  < 1  ,\n  ID\n.EQ."lk")').records[0]
    assert str(record.ignore[0]) == 'WGT  < 1', 'ID\n.EQ."lk"'

    record = parser.parse('$DATA      ../pheno.dta IGNORE=@ IGNORE(APGR.GT.23)\n').records[0]
    record = record.remove_ignore().remove_accept()
    assert str(record) == '$DATA      ../pheno.dta IGNORE=@ \n'


def test_comments(parser):
    record = parser.parse('$DATA pheno.dta IGNORE=@;MYCOMMENT').records[0]
    assert str(record) == '$DATA pheno.dta IGNORE=@;MYCOMMENT'


def test_data_infile(parser):
    record = parser.parse('$INFILE pheno.dta').records[0]
    assert record.name == 'DATA'
    assert record.filename == 'pheno.dta'
    assert record.raw_name == '$INFILE'


def test_comment(parser):
    contents = r"""$DATA     cpt7.dta IGNORE= #
;         Dataset
"""
    record = parser.parse(contents).records[0]
    record = record.set_ignore_character("A")
    assert str(record) == '$DATA     cpt7.dta \n;         Dataset\nIGNORE=A\n'



from ..model_database import NullModelDatabase
from .baseclass import ToolDatabase


class NullToolDatabase(ToolDatabase):
    """Dummy tool database

    No operation does anything. This database can be used if no storing of files
    is desirable.
    """

    def __init__(self, toolname, **kwargs):
        super().__init__(toolname)
        self.model_database = NullModelDatabase()

    def store_local_file(self, source_path):
        pass

    def store_results(self, res):
        pass

    def store_metadata(self, metadata):
        pass

    def read_metadata(self):
        pass



import pytest

from pharmpy.tools.retries.tool import create_new_parameter_inits


@pytest.mark.parametrize(
    ('scale',),
    (('UCP',), ('normal',)),
)
def test_parameter_inits(pheno, scale):
    fraction = 0.1
    seed = 13
    new_parameters = create_new_parameter_inits(pheno, fraction, scale, seed)

    for parameter, value in new_parameters.items():
        param_init = pheno.parameters[parameter].init
        upper = param_init + fraction * param_init
        lower = param_init - fraction * param_init
        assert lower < value < upper



import json
import os
import os.path
import shutil
import subprocess
import time
import uuid
import warnings
from itertools import repeat
from pathlib import Path

from pharmpy.model.external.nonmem import convert_model
from pharmpy.modeling import write_csv, write_model
from pharmpy.tools.external.nonmem import conf, parse_modelfit_results
from pharmpy.workflows import ModelEntry

PARENT_DIR = f'..{os.path.sep}'


def execute_model(model_entry, db):
    assert isinstance(model_entry, ModelEntry)
    model = model_entry.model

    database = db.model_database
    parent_model = model.parent_model
    model = convert_model(model)
    model = model.replace(parent_model=parent_model)
    path = Path.cwd() / f'NONMEM_run_{model.name}-{uuid.uuid1()}'

    # NOTE: This deduplicates the dataset before running NONMEM, so we know which
    # filename to give to this dataset.
    database.store_model(model)
    # NOTE: We cannot reuse model_with_correct_datapath as the model object
    # later because it might have lost some of the ETA names mapping due to the
    # current incomplete implementation of serialization of Pharmpy Model
    # objects through the NONMEM plugin. Hopefully we can get rid of this
    # hack later.
    model_with_correct_datapath = database.retrieve_model(model.name)
    stream = model_with_correct_datapath.internals.control_stream
    data_record = stream.get_records('DATA')[0]
    relative_dataset_path = data_record.filename

    # NOTE: We set up a directory tree that replicates the structure generated by
    # the database so that NONMEM writes down the correct relative paths in
    # generated files such as results.lst.
    # NOTE: It is important that we do this in a DB-agnostic way so that we do
    # not depent on its implementation.
    depth = relative_dataset_path.count(PARENT_DIR)
    # NOTE: This creates an FS tree branch x/x/x/x/...
    model_path = path.joinpath(*repeat('x', depth))
    meta = model_path / '.pharmpy'
    meta.mkdir(parents=True, exist_ok=True)
    # NOTE: This removes the leading ../
    relative_dataset_path_suffix = relative_dataset_path[len(PARENT_DIR) * depth :]
    # NOTE: We do not support non-leading ../, e.g. a/b/../c
    assert PARENT_DIR not in relative_dataset_path_suffix
    dataset_path = path / Path(relative_dataset_path_suffix)
    datasets_path = dataset_path.parent
    datasets_path.mkdir(parents=True, exist_ok=True)

    # NOTE: Write dataset and model files so they can be used by NONMEM.
    model = write_csv(model, path=dataset_path, force=True)
    model = write_model(model, path=model_path, force=True)

    args = nmfe(
        model.name + model.filename_extension,
        'results.lst',
    )

    stdout = model_path / 'stdout'
    stderr = model_path / 'stderr'

    with open(stdout, "wb") as out, open(stderr, "wb") as err:
        result = subprocess.run(
            args, stdin=subprocess.DEVNULL, stderr=err, stdout=out, cwd=str(model_path)
        )

    basename = Path(model.name)

    results_path = model_path / 'results.lst'
    start = time.time()
    timeout = 5

    while True:
        try:
            results_path.rename((model_path / basename).with_suffix('.lst'))
            break
        except FileNotFoundError:
            elapsed_time = time.time() - start
            if elapsed_time >= timeout:
                warnings.warn(f'UNEXPECTED Could not find .lst-file after waiting {elapsed_time}s')
                break
            else:
                time.sleep(1)

    metadata = {
        'plugin': 'nonmem',
        'path': str(model_path),
    }

    plugin = {
        'commands': [
            {
                'args': args,
                'returncode': result.returncode,
                'stdout': 'stdout',
                'stderr': 'stderr',
            }
        ]
    }

    modelfit_results = parse_modelfit_results(model, model_path / basename)

    log = modelfit_results.log if modelfit_results else None
    model_entry = model_entry.attach_results(modelfit_results=modelfit_results, log=log)

    with database.transaction(model_entry) as txn:
        if (
            not (model_path / basename).with_suffix('.lst').is_file()
            or not (model_path / basename).with_suffix('.ext').is_file()
        ):
            warnings.warn(f'Expected result files do not exist, copying everything: {basename}')
            for file in path.glob('*'):
                txn.store_local_file(file)
        else:
            for suffix in ['.lst', '.ext', '.phi', '.cov', '.cor', '.coi', '.grd']:
                file_path = (model_path / basename).with_suffix(suffix)
                txn.store_local_file(file_path)

            for rec in model.internals.control_stream.get_records('TABLE'):
                txn.store_local_file(model_path / rec.path)

        txn.store_local_file(stdout)
        txn.store_local_file(stderr)

        plugin_path = model_path / 'nonmem.json'
        with open(plugin_path, 'w') as f:
            json.dump(plugin, f, indent=2)

        txn.store_local_file(plugin_path)

        txn.store_metadata(metadata)

        txn.store_model_entry()

    return model_entry


def nmfe_path():
    if os.name == 'nt':
        nmfe_candidates = ['nmfe75.bat', 'nmfe74.bat', 'nmfe73.bat']
    else:
        nmfe_candidates = ['nmfe75', 'nmfe74', 'nmfe73']
    path = conf.default_nonmem_path
    if path != Path(''):
        path /= 'run'
        for nmfe in nmfe_candidates:
            candidate_path = path / nmfe
            if candidate_path.is_file():
                path = candidate_path
                break
        else:
            raise FileNotFoundError(f'Cannot find nmfe script for NONMEM ({path})')
    else:
        # Not in configuration file
        for nmfe in nmfe_candidates:
            candidate_path = shutil.which(nmfe)
            if candidate_path is not None:
                path = candidate_path
                break
        else:
            raise FileNotFoundError(
                'No path to NONMEM configured in pharmpy.conf and nmfe not in PATH'
            )
    return str(path)


def nmfe(*args):
    conf_args = []
    if conf.licfile is not None:
        conf_args.append(f'-licfile={str(conf.licfile)}')

    return [
        nmfe_path(),
        *args,
        *conf_args,
    ]



from pharmpy.modeling import (
    plot_abs_cwres_vs_ipred,
    plot_cwres_vs_idv,
    plot_dv_vs_ipred,
    plot_dv_vs_pred,
    plot_individual_predictions,
    plot_iofv_vs_iofv,
    plot_transformed_eta_distributions,
)
from pharmpy.tools import read_modelfit_results


def test_plot_iofv_vs_iofv(testdata):
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_real.mod')
    iofv = res.individual_ofv
    assert plot_iofv_vs_iofv(iofv, iofv, "run1", "run1")


def test_plot_individual_predictions(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_real.mod')
    plot = plot_individual_predictions(model, res.predictions)
    assert plot
    plot = plot_individual_predictions(model, res.predictions[['PRED']], individuals=[1, 2, 5])
    assert plot


def test_plot_transformed_eta_distributions(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'qa' / 'boxcox.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'qa' / 'boxcox.mod')
    pe = res.parameter_estimates
    ie = res.individual_estimates
    plot = plot_transformed_eta_distributions(model, pe, ie)
    assert plot


def test_plot_dv_vs_ipred(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_real.mod')
    plot = plot_dv_vs_ipred(model, res.predictions)
    assert plot


def test_plot_dv_vs_pred(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_real.mod')
    plot = plot_dv_vs_pred(model, res.predictions)
    assert plot


def test_plot_cwres_vs_idv(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_real.mod')
    plot = plot_cwres_vs_idv(model, res.residuals)
    assert plot


def test_plot_abs_cwres_vs_ipred(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_real.mod')
    plot = plot_abs_cwres_vs_ipred(model, res.predictions, res.residuals)
    assert plot



import shutil

import pytest
from pandas.testing import assert_frame_equal

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Model
from pharmpy.tools import fit, read_modelfit_results, resume_tool, retrieve_models, run_tool


def test_run_tool_ruvsearch_resume_flag(tmp_path, testdata):
    with chdir(tmp_path):
        for path in (testdata / 'nonmem').glob('pheno_real.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'sdtab1', tmp_path)
        model = Model.parse_model('pheno_real.mod')
        results = read_modelfit_results('pheno_real.mod')
        model = model.replace(datainfo=model.datainfo.replace(path=tmp_path / 'pheno.dta'))
        path = 'x'
        for i, resume in enumerate([False, False, True]):
            try:
                res = run_tool(
                    'ruvsearch',
                    model=model,
                    results=results,
                    groups=4,
                    p_value=0.05,
                    skip=[],
                    path=path,
                    resume=resume,
                )
                if i != 0 and not resume:
                    assert False
            except FileExistsError as e:
                if i == 0 or resume:
                    raise e

            if i == 0 or resume:
                assert res


def test_run_tool_iivsearch_resume_flag(tmp_path, testdata, model_count):
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox_simulated_normal.csv', tmp_path)
        model_start = Model.parse_model('mox2.mod')
        model_start = model_start.replace(
            datainfo=model_start.datainfo.replace(path=tmp_path / 'mox_simulated_normal.csv')
        )
        start_res = fit(model_start)
        # FIXME: Remove
        model_start = model_start.replace(modelfit_results=start_res)

        path = 'x'
        for i, resume in enumerate([False, False, True]):
            try:
                res = run_tool(
                    'iivsearch',
                    'brute_force_no_of_etas',
                    model=model_start,
                    results=start_res,
                    path=path,
                    resume=resume,
                )
                if i != 0 and not resume:
                    assert False
            except FileExistsError as e:
                if i == 0 or resume:
                    raise e

            if i == 0 or resume:
                no_of_candidate_models = 7
                assert len(res.summary_tool) == no_of_candidate_models + 1
                assert len(res.summary_models) == no_of_candidate_models + 1

                res_models = [
                    model for model in retrieve_models(res) if model.name != 'input_model'
                ]
                assert len(res_models) == no_of_candidate_models

                # assert res_models[-1].modelfit_results

                rundir = tmp_path / path
                assert rundir.is_dir()
                # Candidate model directories + input_model directory + results.json + results.csv
                assert len(list((rundir / 'models').iterdir())) == no_of_candidate_models + 3
                assert model_count(rundir) == no_of_candidate_models
                assert (rundir / 'metadata.json').exists()


@pytest.mark.parametrize(
    'search_space, no_of_models, last_model_parent_name',
    [
        ('ABSORPTION([FO,ZO]);PERIPHERALS([0,1])', 4, 'modelsearch_run2'),
    ],
)
def test_run_tool_modelsearch_resume_flag(
    tmp_path, testdata, model_count, search_space, no_of_models, last_model_parent_name
):
    shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
    shutil.copy2(testdata / 'nonmem' / 'models' / 'mox_simulated_normal.csv', tmp_path)
    # FIXME: Temporary workaround so that read in parameter estimates use the Pharmpy name
    with chdir(tmp_path):
        model_start = Model.parse_model('mox2.mod')
        model_start = model_start.replace(
            datainfo=model_start.datainfo.replace(path=tmp_path / 'mox_simulated_normal.csv')
        )

        start_res = fit(model_start)
        model_start = model_start.replace(modelfit_results=start_res)

        path = 'x'
        for i, resume in enumerate([False, False, True]):
            try:
                res = run_tool(
                    'modelsearch',
                    search_space,
                    'exhaustive_stepwise',
                    model=model_start,
                    results=model_start.modelfit_results,
                    path=path,
                    resume=resume,
                )
                if i != 0 and not resume:
                    assert False
            except FileExistsError as e:
                if i == 0 or resume:
                    raise e

            if i == 0 or resume:
                assert len(res.summary_tool) == no_of_models + 1
                assert len(res.summary_models) == no_of_models + 1
                assert len(res.models) == no_of_models + 1
                assert res.models[-1].modelfit_results

                assert res.models[1].parent_model == 'mox2'
                assert res.models[-1].parent_model == last_model_parent_name
                if last_model_parent_name != 'mox2':
                    last_model_features = res.summary_tool.loc[res.models[-1].name]['description']
                    parent_model_features = res.summary_tool.loc[last_model_parent_name][
                        'description'
                    ]
                    assert (
                        last_model_features[: len(parent_model_features)] == parent_model_features
                    )

                rundir = tmp_path / path
                assert rundir.is_dir()
                # Candidate model directories + input_model directory + results.json + results.csv
                assert len(list((rundir / 'models').iterdir())) == no_of_models + 3
                assert model_count(rundir) == no_of_models
                assert (rundir / 'results.json').exists()
                assert (rundir / 'results.csv').exists()
                assert (rundir / 'metadata.json').exists()


def test_resume_tool_ruvsearch(tmp_path, testdata):
    with chdir(tmp_path):
        for path in (testdata / 'nonmem').glob('pheno_real.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'sdtab1', tmp_path)

        model = Model.parse_model('pheno_real.mod')
        results = read_modelfit_results('pheno_real.mod')
        model = model.replace(datainfo=model.datainfo.replace(path=tmp_path / 'pheno.dta'))
        path = 'x'
        run_tool_res = run_tool(
            'ruvsearch',
            model=model,
            results=results,
            groups=4,
            p_value=0.05,
            skip=[],
            path=path,
        )
        assert run_tool_res

        resume_tool_res = resume_tool(path)
        assert resume_tool_res

        assert type(resume_tool_res) == type(run_tool_res)  # noqa: E721

        assert_frame_equal(resume_tool_res.cwres_models, run_tool_res.cwres_models)
        assert_frame_equal(resume_tool_res.summary_individuals, run_tool_res.summary_individuals)
        assert_frame_equal(
            resume_tool_res.summary_individuals_count, run_tool_res.summary_individuals_count
        )
        assert resume_tool_res.final_model.name == run_tool_res.final_model.name
        assert_frame_equal(resume_tool_res.summary_models, run_tool_res.summary_models)
        assert_frame_equal(resume_tool_res.summary_tool, run_tool_res.summary_tool)
        assert_frame_equal(resume_tool_res.summary_errors, run_tool_res.summary_errors)
        assert type(resume_tool_res.tool_database) == type(run_tool_res.tool_database)  # noqa: E721
        assert resume_tool_res.tool_database.to_dict() == run_tool_res.tool_database.to_dict()



import json
import shutil
from itertools import count
from pathlib import Path

from pharmpy.internals.fs.path import path_absolute
from pharmpy.model import Model
from pharmpy.workflows.results import ModelfitResults

from ..model_database import LocalModelDirectoryDatabase
from .baseclass import ToolDatabase


class LocalDirectoryToolDatabase(ToolDatabase):
    """ToolDatabase in a local directory

    Parameters
    ----------
    toolname : str
        Name of the tool
    path : str or Path
        Path to directory. Will be created if it does not exist.
    exist_ok : bool
        Whether to allow using an existing database.
    """

    def __init__(self, toolname=None, path=None, exist_ok=False):
        if path is None:
            if toolname is None:
                raise ValueError('Must specify toolname when not specifying path')
            for i in count(1):
                name = f'{toolname}_dir{i}'
                path = Path(name)
                try:
                    path.mkdir(parents=True)
                    break
                except FileExistsError:
                    pass
        else:
            path = Path(path)
            path.mkdir(parents=True, exist_ok=exist_ok)

        assert path is not None
        self.path = path_absolute(path)

        modeldb = LocalModelDirectoryDatabase(self.path / 'models')
        self.model_database = modeldb

        if toolname is None:
            toolname = self.read_metadata()['tool_name']

        super().__init__(toolname)

    def to_dict(self):
        return {'toolname': self.toolname, 'path': str(self.path)}

    @classmethod
    def from_dict(cls, d):
        return cls(**d, exist_ok=True)

    def store_local_file(self, source_path):
        if Path(source_path).is_file():
            shutil.copy2(source_path, self.path)

    def store_results(self, res):
        res.to_json(path=self.path / 'results.json')
        res.to_csv(path=self.path / 'results.csv')

    def store_metadata(self, metadata):
        path = self.path / 'metadata.json'
        with open(path, 'w') as f:
            json.dump(metadata, f, indent=4, cls=MetadataJSONEncoder)

    def read_metadata(self):
        path = self.path / 'metadata.json'
        with open(path, 'r') as f:
            return json.load(f, cls=MetadataJSONDecoder)


class MetadataJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Model):
            # NOTE: This is only used by modelfit at the moment since we encode
            # models for other tools upstream.
            return obj.name
        if isinstance(obj, ModelfitResults):
            return obj.to_json()
        return super().default(obj)


class MetadataJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        return obj



import importlib


def parse_modelfit_results(model, path):
    module_name = model.__class__.__module__
    a = module_name.split(".")
    name = a[3]
    tool_module_name = f'pharmpy.tools.external.{name}'
    module = importlib.import_module(tool_module_name)

    res = module.parse_modelfit_results(model, path)
    return res



from pharmpy.modeling import (
    add_metabolite,
    add_peripheral_compartment,
    has_presystemic_metabolite,
    remove_peripheral_compartment,
)


def test_add_metabolite(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_conc.mod')
    model = add_metabolite(model)
    odes = model.statements.ode_system
    assert odes.compartment_names == ['CENTRAL', 'METABOLITE']
    assert not has_presystemic_metabolite(model)
    a = model.model_code.split('\n')
    assert a[20] == 'IF (DVID.EQ.1) Y = Y'
    assert a[21] == 'IF (DVID.EQ.2) Y = Y_M'

    assert odes.central_compartment.name == 'CENTRAL'

    model = add_peripheral_compartment(model, "METABOLITE")
    odes = model.statements.ode_system

    assert odes.find_peripheral_compartments("METABOLITE")
    assert not odes.find_peripheral_compartments()

    model = add_peripheral_compartment(model)
    odes = model.statements.ode_system

    assert odes.find_peripheral_compartments("METABOLITE")
    assert odes.find_peripheral_compartments()

    model = remove_peripheral_compartment(model, "METABOLITE")
    odes = model.statements.ode_system

    assert not odes.find_peripheral_compartments("METABOLITE")
    assert odes.find_peripheral_compartments()

    model = remove_peripheral_compartment(model)
    odes = model.statements.ode_system

    assert not odes.find_peripheral_compartments("METABOLITE")
    assert not odes.find_peripheral_compartments()


def test_presystemic_metabolite(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_conc.mod')
    model = add_metabolite(model, presystemic=True)
    odes = model.statements.ode_system

    assert odes.compartment_names == ['DEPOT', 'CENTRAL', 'METABOLITE']

    depot = odes.find_depot(model.statements)
    central = odes.central_compartment
    metabolite = odes.find_compartment("METABOLITE")

    assert has_presystemic_metabolite(model)
    assert odes.get_flow(depot, central)
    assert odes.get_flow(depot, metabolite)
    assert odes.get_flow(central, metabolite)

    assert odes.central_compartment.name == 'CENTRAL'

    model = add_peripheral_compartment(model, "METABOLITE")
    odes = model.statements.ode_system

    assert odes.find_peripheral_compartments("METABOLITE")
    assert not odes.find_peripheral_compartments()

    model = add_peripheral_compartment(model)
    odes = model.statements.ode_system

    assert odes.find_peripheral_compartments("METABOLITE")
    assert odes.find_peripheral_compartments()

    model = remove_peripheral_compartment(model, "METABOLITE")
    odes = model.statements.ode_system

    assert not odes.find_peripheral_compartments("METABOLITE")
    assert odes.find_peripheral_compartments()

    model = remove_peripheral_compartment(model)
    odes = model.statements.ode_system

    assert not odes.find_peripheral_compartments("METABOLITE")
    assert not odes.find_peripheral_compartments()



from pharmpy.internals.fs.symlink import create_directory_symlink


def test_create_directory_symlink(tmp_path):
    target_dir = tmp_path / "target"
    target_dir.mkdir()
    link_dir = tmp_path / "link"

    create_directory_symlink(link_dir, target_dir)

    assert link_dir.is_dir()
    assert target_dir.is_dir()

    test_file_path = target_dir / "testfile"
    link_file_path = link_dir / "testfile"
    assert not test_file_path.exists()
    assert not link_file_path.exists()

    epictetus_quote = "No great thing is created suddenly."
    with open(test_file_path, 'w') as fh:
        fh.write(epictetus_quote)

    with open(link_file_path, 'r') as fh:
        text = fh.read()

    assert text == epictetus_quote



import re
from datetime import datetime
from typing import Dict, Optional, Union

import dateutil.parser
from packaging import version

from pharmpy.deps import numpy as np


class NONMEMResultsFile:
    """Representing and parsing a NONMEM results file (aka lst-file)
    This is not a generic output object and will be combined by other classes
    into new structures.
    We rely on tags in NONMEM >= 7.2.0
    """

    def __init__(self, path=None, log=None):
        self.log = log
        self.table = {}
        self.nonmem_version = None
        self.runtime_total = None
        if path is not None:
            for name, content in self.table_blocks(path):
                if name == 'INIT':
                    self.nonmem_version = content.pop('nonmem_version', None)
                elif name == 'runtime':
                    self.runtime_total = content.pop('total', None)
                else:
                    self.table[name] = content

    @property
    def _supported_nonmem_version(self):
        return NONMEMResultsFile.supported_version(self.nonmem_version)

    def estimation_status(self, table_number):
        result = NONMEMResultsFile.unknown_termination()
        if self._supported_nonmem_version:
            if table_number in self.table.keys():
                for key in result.keys():
                    result[key] = self.table[table_number].get(key)
            else:
                result['minimization_successful'] = False
        return result

    def covariance_status(self, table_number):
        result = NONMEMResultsFile.unknown_covariance()
        if self._supported_nonmem_version:
            if table_number in self.table.keys():
                for key in result.keys():
                    result[key] = self.table[table_number].get(key)
            else:
                result['covariance_step_ok'] = False
        return result

    def ofv(self, table_number):
        ofv = None
        if self._supported_nonmem_version:
            if table_number in self.table.keys():
                ofv = self.table[table_number].get('OBJV')
        if ofv is not None:
            ofv = float(ofv)
        else:
            ofv = np.nan
        return ofv

    @staticmethod
    def supported_version(nonmem_version):
        return nonmem_version is not None and version.parse(nonmem_version) >= version.parse(
            '7.2.0'
        )

    @staticmethod
    def unknown_covariance() -> Dict[str, Optional[Union[bool, float]]]:
        return {'covariance_step_ok': None}

    @staticmethod
    def unknown_termination() -> Dict[str, Optional[Union[bool, float]]]:
        return {
            'minimization_successful': None,
            'estimate_near_boundary': None,
            'rounding_errors': None,
            'maxevals_exceeded': None,
            'significant_digits': np.nan,
            'function_evaluations': np.nan,
            'warning': None,
        }

    @staticmethod
    def cleanup_version(v):
        if v == 'V':
            v = '5.0'
        elif v == 'VI':
            v = '6.0'
        return v

    @staticmethod
    def parse_tere(rows):
        result = NONMEMResultsFile.unknown_covariance()
        result['covariance_step_ok'] = False
        result['estimation_runtime'] = np.nan
        if len(rows) < 1:
            return result

        cov_not_ok = re.compile(
            r'( INTERPRET VARIANCE-COVARIANCE OF ESTIMATES WITH CARE)|'
            r'(R|S) MATRIX ALGORITHMICALLY SINGULAR'
        )
        # Need variable whitespace
        cov_ok = re.compile(r' Elapsed (covariance|opt\. design)\s+time in seconds: ')
        est_time = re.compile(r' Elapsed estimation\s+time in seconds:\s+(\d+\.*\d+)')

        for row in rows:
            if cov_not_ok.match(row):
                result['covariance_step_ok'] = False
                break
            if cov_ok.match(row):
                result['covariance_step_ok'] = True
                break
            m = est_time.match(row)
            if m:
                result['estimation_runtime'] = float(m.group(1))
        return result

    @staticmethod
    def parse_termination(rows):
        result = NONMEMResultsFile.unknown_termination()
        if len(rows) < 1:  # Will happen if e.g. TERMINATED BY OBJ during estimation
            result['minimization_successful'] = False
            return result
        result['estimate_near_boundary'] = False
        result['rounding_errors'] = False
        result['maxevals_exceeded'] = False
        result['warning'] = False

        success = [
            re.compile(r'0MINIMIZATION SUCCESSFUL'),
            re.compile(
                r' (REDUCED STOCHASTIC PORTION|OPTIMIZATION|'
                + r'BURN-IN|EXPECTATION ONLY PROCESS)'
                + r'( STATISTICALLY | WAS | )(COMPLETED|NOT TESTED)'
            ),
            re.compile(r'1OBJECTIVE FUNCTION IS TO BE EVALUATED'),
        ]
        failure = [
            re.compile(r'0MINIMIZATION TERMINATED'),
            re.compile(r'0SEARCH WITH ESTIMATION STEP WILL NOT PROCEED'),
            re.compile(r'\s*INDIVIDUAL OBJECTIVE FUNCTION VALUES ARE ALL ZERO\. PROBLEM ENDED'),
        ]
        maybe = re.compile(
            r' (REDUCED STOCHASTIC PORTION|OPTIMIZATION|'
            + r'BURN-IN|EXPECTATION ONLY PROCESS)'
            + r'( WAS | )NOT COMPLETED'
        )  # Success only if next line USER INTERRUPT
        misc = {
            'estimate_near_boundary': re.compile(
                r'0(ESTIMATE OF THETA IS NEAR THE BOUNDARY AND|'
                + r'PARAMETER ESTIMATE IS NEAR ITS BOUNDARY)'
            ),
            'rounding_errors': re.compile(r' DUE TO ROUNDING ERRORS'),
            'maxevals_exceeded': re.compile(r' DUE TO MAX. NO. OF FUNCTION EVALUATIONS EXCEEDED'),
            'warning': re.compile(r' HOWEVER, PROBLEMS OCCURRED WITH THE MINIMIZATION.'),
        }
        sig_digits = re.compile(r' NO. OF SIG. DIGITS IN FINAL EST.:\s*(\S+)')  # Only classical est
        sig_digits_unreport = re.compile(
            r'\w*(NO. OF SIG. DIGITS UNREPORTABLE)\w*\n'
        )  # Only classical est
        feval = re.compile(r' NO. OF FUNCTION EVALUATIONS USED:\s*(\S+)')  # Only classical est
        ofv_with_constant = re.compile(r' OBJECTIVE FUNCTION VALUE WITH CONSTANT:\s*(\S+)')

        maybe_success = False
        for row in rows:
            if maybe_success:
                result['minimization_successful'] = bool(re.search(r'USER INTERRUPT', row))
                break
            for p in success:
                if p.match(row):
                    result['minimization_successful'] = True
                    break
            if result['minimization_successful'] is not None:
                break
            for p in failure:
                if p.match(row):
                    result['minimization_successful'] = False
                    break
            if result['minimization_successful'] is not None:
                break
            maybe_success = bool(maybe.match(row))
        for row in rows:
            m = sig_digits.match(row)
            if m:
                result['significant_digits'] = float(m.group(1))
                continue
            m = sig_digits_unreport.match(row)
            if m:
                result['significant_digits'] = np.nan
            m = ofv_with_constant.match(row)
            if m:
                result['ofv_with_constant'] = float(m.group(1))
                continue
            m = feval.match(row)
            if m:
                result['function_evaluations'] = int(m.group(1))
                continue
            for name, p in misc.items():
                if p.match(row):
                    result[name] = True
                    break
        return result

    @staticmethod
    def parse_runtime(row, row_next=None):
        # TODO: Support AM/PM
        weekday_month_en = re.compile(
            r'^\s*(Sun|Mon|Tue|Wed|Thu|Fri|Sat)'
            r'\s+(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)'  # Month
            r'\s+(\d+)'  # Day
            r'\s+.*'
            r'\s+(\d{4})'  # Year
        )
        weekday_month_sv = re.compile(
            r'^\s*(mån|tis|ons|tor|fre|lör|sön)'
            r'\s+(\d+)'  # Day
            r'\s+(jan|feb|mar|apr|maj|jun|jul|aug|sep|okt|nov|dec)'  # Month
            r'\s+(\d+)'  # Year
        )
        day_month_year = re.compile(r'^(\d{2})/(\d{2})/(\d{4})\s*$')  # dd/mm/yyyy
        year_month_day = re.compile(r'^(\d{4})-\d{2}-\d{2}\s*$')  # yyyy/mm/dd
        timestamp = re.compile(r'([0-9]{2}:[0-9]{2}:[0-9]{2})')

        month_no = {
            'JAN': 1,
            'FEB': 2,
            'MAR': 3,
            'APR': 4,
            'MAY': 5,
            'JUN': 6,
            'JUL': 7,
            'AUG': 8,
            'SEP': 9,
            'OCT': 10,
            'NOV': 11,
            'DEC': 12,
        }
        month_trans = {'MAJ': 'MAY', 'OKT': 'OCT'}

        def _dmy(row):
            if match_en := weekday_month_en.match(row):
                _, month, day, year = match_en.groups()
                return day, month, year

            elif match_sv := weekday_month_sv.match(row):
                _, day, month, year = match_sv.groups()
                return day, month, year

            return None

        dmy = _dmy(row)
        if dmy is not None:
            day, month, year = dmy
            try:
                month = month_no[month.upper()]
            except KeyError:
                month_en = month_trans[month.upper()]
                month = month_no[month_en.upper()]

            date = datetime(int(year), int(month), int(day))

            match = timestamp.search(row)
            if match is None:
                return date

            time_str = match.groups()[0]
            time = dateutil.parser.parse(time_str).time()
            return datetime.combine(date, time)

        elif (match_day_first := day_month_year.match(row)) or year_month_day.match(row):
            date = dateutil.parser.parse(row, dayfirst=bool(match_day_first))

            if row_next is None:
                return date

            time = dateutil.parser.parse(row_next).time()
            return datetime.combine(date, time)

    def log_items(self, lines):
        fulltext = '\n'.join(lines)

        warnings = []
        errors = []

        warning_patterns = [
            re.compile(r'0WARNING:((\s+.+\n)+)'),
            re.compile(r'0(PARAMETER ESTIMATE IS NEAR ITS BOUNDARY)'),
            re.compile(r'0(MINIMIZATION SUCCESSFUL\n\s*HOWEVER.+\n)'),
        ]

        for pattern in warning_patterns:
            match = pattern.search(fulltext)
            if match:
                message = match.group(1)
                message_split = message.split('\n')
                message_trimmed = '\n'.join([m.strip() for m in message_split])
                warnings.append(message_trimmed.strip())

        error_patterns = [
            re.compile(
                r'(AN ERROR WAS FOUND IN THE CONTROL STATEMENTS\.(.*\n)+'
                r'.+UPPER OR LOWER BOUNDS\.)'
            ),
            re.compile(
                r'(INITIAL ESTIMATE OF OMEGA HAS A NONZERO BLOCK WHICH IS NUMERICALLY NOT '
                r'POSITIVE DEFINITE)'
            ),
            re.compile(r'0(UPPER BOUNDS INAPPROPRIATE)'),
            re.compile(r'0(PRED EXIT CODE = 1\n(.*\n)+.+MAY BE TOO LARGE\.)'),
            re.compile(
                r'0(PRED EXIT CODE = 1\n(.*\n)+\s*'
                r'NUMERICAL DIFFICULTIES OBTAINING THE SOLUTION\.)\s*\n'
            ),
            re.compile(r'0(PRED EXIT CODE = 1\n(.*\n)+\s+.+IS TOO CLOSE TO AN EIGENVALUE\s*)\n'),
            re.compile(r'0(PRED EXIT CODE = 1\n(.*\n)+\s+.+IS VERY LARGE\.\s*)\n'),
            re.compile(r'0(PROGRAM TERMINATED BY OBJ\n\s*MESSAGE ISSUED FROM ESTIMATION STEP)'),
            re.compile(
                r'0(PROGRAM TERMINATED BY OBJ\n(.*\n)*\s*'
                r'MESSAGE ISSUED FROM ESTIMATION STEP\n\s*'
                r'((AT 0TH ITERATION, UPON EVALUATION OF GRADIENT.*)|'
                r'(AT INITIAL OBJ. FUNCTION EVALUATION)))\n'
            ),
            re.compile(r'0(MINIMIZATION TERMINATED\n\s*DUE TO ROUNDING ERRORS.+)\n'),
            re.compile(r'0(MINIMIZATION TERMINATED\n\s*DUE TO ZERO GRADIENT)\n'),
            re.compile(
                r'0(MINIMIZATION TERMINATED\n\s*DUE TO MAX. NO. OF FUNCTION EVALUATIONS EXCEEDED)\n'
            ),
            re.compile(r'0(MINIMIZATION TERMINATED\n(.*\n)+\s*IS NON POSITIVE DEFINITE)\n'),
            re.compile(
                r'0(MINIMIZATION TERMINATED\n(.*\n)+\s*SUM OF "SQUARED" WEIGHTED INDIVIDUAL '
                r'RESIDUALS IS INFINITE)\n'
            ),
            re.compile(r'\s*(NO. OF SIG. DIGITS UNREPORTABLE)\s*\n'),
        ]

        for pattern in error_patterns:
            match = pattern.search(fulltext)
            if match:
                message = match.group(1)
                message_split = message.split('\n')
                message_trimmed = '\n'.join([m.strip() for m in message_split])
                errors.append(message_trimmed.strip())

        if self.log is not None:
            for message in warnings:
                self.log = self.log.log_warning(message)
            for message in errors:
                self.log = self.log.log_error(message)

    def tag_items(self, path):
        nmversion = re.compile(r'1NONLINEAR MIXED EFFECTS MODEL PROGRAM \(NONMEM\) VERSION\s+(\S+)')
        tag = re.compile(r'\s*#([A-Z]{4}):\s*(.*)')
        end_TERE = re.compile(
            r'(0|1)'
        )  # The part we need after #TERE: will precede the next ^1 or ^0
        cleanup = re.compile(r'\*+\s*')
        TERM = []
        TERE = []
        found_TERM = False
        found_TERE = False
        runtime = None
        endtime_index = None

        with open(path, 'rb') as fp:
            binary = fp.readlines()
            # Since lst-files sometimes can have mixed encodings:
            # Allow for separate encodings of date strings and the rest of the content
            # Always try utf-8 first and fallback to latin-1 separately for each section
            try:
                line1 = binary[0].decode('utf-8')
            except UnicodeDecodeError:
                line1 = binary[0].decode('latin-1', errors='ignore')

            try:
                last_line = binary[-1].decode('utf-8')
            except UnicodeDecodeError:
                last_line = binary[-1].decode('latin-1', errors='ignore')

            chunk = b''.join(binary[1:-1])
            try:
                decoded_chunk = chunk.decode('utf-8')
            except UnicodeDecodeError:
                decoded_chunk = chunk.decode('latin-1', errors='ignore')

            lines = [line1]
            lines += decoded_chunk.replace('\r', '').split('\n')
            lines[-1] = last_line  # Replace since lst-files always end with \n

        version_number = None
        starttime = NONMEMResultsFile.parse_runtime(lines[0], lines[1])
        for row in lines:
            m = nmversion.match(row)
            if m:
                version_number = NONMEMResultsFile.cleanup_version(m.group(1))
                yield ('nonmem_version', version_number)
                break  # We will stay at current file position

        self.log_items(lines)

        if NONMEMResultsFile.supported_version(version_number):
            for i, row in enumerate(lines):
                row = row.rstrip()
                m = tag.match(row)
                if m:
                    if m.group(1) == 'TERM':
                        if found_TERM:
                            raise NotImplementedError('Two TERM tags without TERE in between')
                        found_TERM = True
                        TERM = []
                    elif m.group(1) == 'TERE':
                        if not found_TERM:
                            raise NotImplementedError('TERE tag without TERM tag')
                        found_TERE = True
                        yield ('TERM', NONMEMResultsFile.parse_termination(TERM))
                        found_TERM = False
                        TERM = []
                    elif found_TERE:
                        found_TERE = False
                        # Raise NotImplementedError('TERE tag without ^1 or ^0 before next tag')
                    else:
                        v = cleanup.sub('', m.group(2))
                        yield (m.group(1), v.strip())
                elif found_TERE:
                    if end_TERE.match(row):
                        yield ('TERE', NONMEMResultsFile.parse_tere(TERE))
                        found_TERE = False
                        TERE = []
                    else:
                        TERE.append(row)
                elif found_TERM:
                    TERM.append(row)
                if row == 'Stop Time:':
                    endtime_index = i + 1

            if endtime_index is not None:
                second_line = lines[i] if (i := endtime_index + 1) < len(lines) else None
                endtime = NONMEMResultsFile.parse_runtime(lines[endtime_index], second_line)
                if starttime and endtime:
                    runtime = (endtime - starttime).total_seconds()

            if found_TERM:
                yield ('TERM', NONMEMResultsFile.parse_termination(TERM))
            if found_TERE:
                yield ('TERE', NONMEMResultsFile.parse_tere(TERE))
            if runtime is not None:
                yield ('runtime', runtime)

    def table_blocks(self, path):
        block = {}
        table_number = 'INIT'
        for name, content in self.tag_items(path):
            if name == 'TERM' or name == 'TERE':
                for k, v in content.items():
                    block[k] = v
            elif name == 'TBLN':
                if bool(block):
                    yield (table_number, block)
                block = {}
                table_number = int(content)
            elif name == 'runtime':
                yield ('runtime', {'total': content})
            else:
                # If already set then it means TBLN was missing, probably $SIM, skip
                if name not in block.keys():
                    block[name] = content
        if bool(block):
            yield (table_number, block)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

from pharmpy.deps import pandas as pd
from pharmpy.model import Model
from pharmpy.modeling import plot_iofv_vs_iofv
from pharmpy.tools import read_modelfit_results
from pharmpy.workflows import Results


@dataclass(frozen=True)
class LinearizeResults(Results):
    ofv: Optional[float] = None
    iofv: Optional[Any] = None
    iofv_plot: Optional[Any] = None


def calculate_results(base_model, base_model_results, linear_model, linear_model_results):
    iofv = pd.DataFrame(
        {
            'base': base_model_results.individual_ofv,
            'linear': linear_model_results.individual_ofv,
            'delta': linear_model_results.individual_ofv - base_model_results.individual_ofv,
        }
    )
    ofv = pd.DataFrame(
        {
            'ofv': [
                base_model_results.ofv,
                linear_model_results.ofv_iterations.iloc[0],
                linear_model_results.ofv,
            ]
        },
        index=['base', 'lin_evaluated', 'lin_estimated'],
    )
    iofv1 = base_model_results.individual_ofv
    iofv2 = linear_model_results.individual_ofv
    iofv_plot = plot_iofv_vs_iofv(iofv1, iofv2, base_model.name, linear_model.name)
    res = LinearizeResults(ofv=ofv, iofv=iofv, iofv_plot=iofv_plot)
    return res


def psn_linearize_results(path):
    """Create linearize results from a PsN linearize run

    :param path: Path to PsN linearize run directory
    :return: A :class:`QAResults` object
    """
    path = Path(path)

    base_model_path = path / 'scm_dir1' / 'derivatives.mod'
    base_model = Model.parse_model(base_model_path)
    base_model_results = read_modelfit_results(base_model_path)
    lin_path = list(path.glob('*_linbase.mod'))[0]
    lin_model = Model.parse_model(lin_path)
    lin_model_results = read_modelfit_results(lin_path)

    res = calculate_results(base_model, base_model_results, lin_model, lin_model_results)
    return res



import shutil

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Model
from pharmpy.modeling import fix_parameters
from pharmpy.tools import fit, run_iovsearch


def test_default_mox2(tmp_path, model_count, start_modelres):
    with chdir(tmp_path):
        res = run_iovsearch(
            'VISI', rank_type='bic', results=start_modelres[1], model=start_modelres[0]
        )
        rundir = tmp_path / 'iovsearch_dir1'
        assert model_count(rundir) == 8

        assert res.final_model.name == 'iovsearch_run7'


def test_ignore_fixed_iiv(tmp_path, model_count, start_modelres):
    with chdir(tmp_path):
        start_model = fix_parameters(start_modelres[0], 'IIV_CL')
        res = run_iovsearch('VISI', results=start_modelres[1], model=start_model)
        assert len(res.summary_models) == 5


def test_rank_type_ofv_mox2(tmp_path, model_count, start_modelres):
    with chdir(tmp_path):
        res = run_iovsearch(
            'VISI', results=start_modelres[1], model=start_modelres[0], rank_type='ofv'
        )
        rundir = tmp_path / 'iovsearch_dir1'
        assert model_count(rundir) == 8

        assert res.final_model.name == 'iovsearch_run7'


def test_default_mox1(tmp_path, model_count, testdata):
    shutil.copy2(testdata / 'nonmem' / 'models' / 'mox1.mod', tmp_path)
    shutil.copy2(testdata / 'nonmem' / 'models' / 'mox_simulated_log.csv', tmp_path)
    with chdir(tmp_path):
        start_model = Model.parse_model('mox1.mod')
        start_res = fit(start_model)
        res = run_iovsearch('VISI', results=start_res, model=start_model)
        rundir = tmp_path / 'iovsearch_dir1'
        assert model_count(rundir) == 7

        assert res.final_model.name == start_model.name



import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.tools import retrieve_models, run_estmethod


@pytest.mark.parametrize(
    'algorithm, methods, parameter_uncertainty_methods, no_of_candidates, advan_ref',
    [
        ('exhaustive', ['FOCE', 'IMP'], None, 2, 'ADVAN2'),
        ('exhaustive_only_eval', ['FOCE', 'IMP'], None, 2, 'ADVAN2'),
        ('exhaustive', ['FOCE'], ['SANDWICH', 'CPG'], 2, 'ADVAN2'),
        ('exhaustive_with_update', ['FOCE'], ['SANDWICH', 'CPG'], 4, 'ADVAN2'),
        ('exhaustive_with_update', ['IMP'], ['SANDWICH', 'CPG'], 5, 'ADVAN2'),
    ],
)
def test_estmethod(
    tmp_path,
    start_modelres,
    model_count,
    testdata,
    algorithm,
    methods,
    parameter_uncertainty_methods,
    no_of_candidates,
    advan_ref,
):
    with chdir(tmp_path):
        res = run_estmethod(
            algorithm,
            methods=methods,
            parameter_uncertainty_methods=parameter_uncertainty_methods,
            model=start_modelres[0],
            results=start_modelres[1],
        )

        assert len(res.summary_tool) == no_of_candidates + 1
        res_models = [model for model in retrieve_models(res) if model.name != 'input_model']
        assert len(res_models) == no_of_candidates
        assert advan_ref in res_models[-1].model_code
        rundir = tmp_path / 'estmethod_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == no_of_candidates
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()
        assert (rundir / 'results.html').exists()



import math
from typing import Dict, Union

from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.model import Model


class UCPScale:
    def __init__(self, theta, omega, sigma, lb, range_ul):
        self.theta = theta
        self.omega = omega
        self.sigma = sigma
        self.lb = lb
        self.range_ul = range_ul

    def __repr__(self):
        return "<UCPScale object>"


def calculate_ucp_scale(model: Model):
    """Calculate a scale for unconstrained parameters for a model

    The UCPScale object can be used to calculate unconstrained parameters
    back into the normal parameter space.

    Parameters
    ----------
    model : Model
        Model for which to calculate an ucp scale

    Returns
    -------
    UCPScale
        A scale object

    See Also
    --------
    calculate_parameters_from_ucp : Calculate parameters from ucp:s

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> scale = calculate_ucp_scale(model)

    """
    omega_symbolic = model.random_variables.etas.covariance_matrix
    omega = omega_symbolic.subs(model.parameters.inits)
    omega = np.array(omega).astype(np.float64)
    scale_omega = _scale_matrix(omega)

    sigma_symbolic = model.random_variables.epsilons.covariance_matrix
    sigma = sigma_symbolic.subs(model.parameters.inits)
    sigma = np.array(sigma).astype(np.float64)
    scale_sigma = _scale_matrix(sigma)

    theta = []
    lb = []
    range_ul_vec = []
    for p in model.parameters:
        if not p.fix:
            if p.symbol not in model.random_variables.free_symbols:
                upper = p.upper if p.upper < 1000000 else 1000000
                lower = p.lower if p.lower > -1000000 else -1000000
                range_ul = upper - lower
                range_prop = (p.init - lower) / range_ul
                scaled = 0.1 - math.log(range_prop / (1.0 - range_prop))
                theta.append(scaled)
                lb.append(lower)
                range_ul_vec.append(range_ul)

    return UCPScale(np.array(theta), scale_omega, scale_sigma, np.array(lb), np.array(range_ul_vec))


def _scale_matrix(A):
    chol = np.linalg.cholesky(A)
    M1 = np.tril(chol)
    v1 = np.diag(M1)
    v2 = v1 / np.exp(0.1)
    M2 = np.diag(v1)
    M3 = np.diag(v2)
    m_scale = np.abs(10 * (M1 - M2)) + M3
    # Convert lower triangle to symmetric
    irows, icols = np.triu_indices(len(m_scale), 1)
    m_scale[irows, icols] = m_scale[icols, irows]
    return m_scale


def calculate_parameters_from_ucp(
    model: Model, scale: UCPScale, ucps: Union[pd.Series, Dict[str, float]]
):
    """Scale parameter values from ucp to normal scale

    Parameters
    ----------
    model : Model
        Pharmpy model
    scale : UCPScale
        A parameter scale
    ucps : pd.Series or dict
        Series of parameter values

    Returns
    -------
    pd.Series
        Parameters on the normal scale

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> scale = calculate_ucp_scale(model)
    >>> values = {'PTVCL': 0.1, 'PTVV': 0.1, 'THETA_3': 0.1, \
                  'IVCL': 0.1, 'IVV': 0.1, 'SIGMA_1_1': 0.1}
    >>> calculate_parameters_from_ucp(model, scale, values)
    PTVCL                    0.004693
    PTVV                      1.00916
    THETA_3                       0.1
    IVCL                    0.0309626
    IVV          0.031127999999999996
    SIGMA_1_1    0.013241000000000001
    dtype: object

    See also
    --------
    calculate_ucp_scale : Calculate the scale for conversion from ucps
    """
    omega_symbolic = model.random_variables.etas.covariance_matrix
    omega = omega_symbolic.subs(dict(ucps))
    omega = np.array(omega).astype(np.float64)
    descaled_omega = _descale_matrix(omega, scale.omega)
    omega_dict = {}
    for symb, numb in zip(omega_symbolic, np.nditer(descaled_omega)):
        if symb.is_Symbol:
            omega_dict[symb] = numb

    sigma_symbolic = model.random_variables.epsilons.covariance_matrix
    sigma = sigma_symbolic.subs(dict(ucps))
    sigma = np.array(sigma).astype(np.float64)
    descaled_sigma = _descale_matrix(sigma, scale.sigma)
    sigma_dict = {}
    for symb, numb in zip(sigma_symbolic, np.nditer(descaled_sigma)):
        if symb.is_Symbol:
            sigma_dict[symb] = numb

    theta = []
    for p in model.parameters:
        if not p.fix:
            if p.symbol not in model.random_variables.free_symbols:
                theta.append(ucps[p.name])
    theta = np.array(theta)

    diff_scale = theta - scale.theta
    prop_scale = np.exp(diff_scale) / (1.0 + np.exp(diff_scale))
    descaled = prop_scale * scale.range_ul + scale.lb

    d = {}
    i = 0
    for p in model.parameters:
        if not p.fix:
            if p.symbol not in model.random_variables.free_symbols:
                d[p.name] = descaled[i]
                i += 1
            elif p.symbol in omega_symbolic.free_symbols:
                d[p.name] = omega_dict[p.symbol]
            else:
                d[p.name] = sigma_dict[p.symbol]

    return pd.Series(d)


def _descale_matrix(A, S):
    exp_diag = np.exp(np.diag(A))
    M = A.copy()
    np.fill_diagonal(M, exp_diag)
    M2 = M * S
    M3 = np.tril(M2)
    return M3 @ M3.T



"""
NONMEM data record class.
"""

from pharmpy.internals.parse import AttrToken, AttrTree
from pharmpy.model import ModelSyntaxError

from .option_record import OptionRecord

TYPES_OF_SPACE = frozenset(('WS', 'NEWLINE'))
TYPES_OF_KEEP = frozenset(('WS', 'NEWLINE', 'COMMENT'))


class DataRecord(OptionRecord):
    @property
    def filename(self):
        """The (raw, unresolved) path of the dataset."""
        filename = self.root.subtree('filename')
        if filename.find('FILENAME'):
            return str(filename)
        else:  # 'QFILENAME'
            return str(filename)[1:-1]

    def set_filename(self, value):
        if not value:
            # erase and replace by * (for previous subproblem)
            new = [AttrToken('ASTERISK', '*')]
            nodes = []
            for child in self.root.children:
                if new and child.rule in TYPES_OF_SPACE:
                    nodes += [child, new.pop()]
                elif child.rule in TYPES_OF_KEEP:
                    nodes += [child]
            root = AttrTree.create('root', nodes)
        else:
            # replace only 'filename' rule and quote appropriately if, but only if, needed
            filename = str(value)
            quoted = [
                '"',
                "'",
                ',',
                ';',
                '(',
                ')',
                '=',
                ' ',
                'IGNORE',
                'NULL',
                'ACCEPT',
                'NOWIDE',
                'WIDE',
                'CHECKOUT',
                'RECORDS',
                'RECORDS',
                'LRECL',
                'NOREWIND',
                'REWIND',
                'NOOPEN',
                'LAST20',
                'TRANSLATE',
                'BLANKOK',
                'MISDAT',
            ]
            if not any(x in filename for x in quoted):
                node = AttrTree.create('filename', {'FILENAME': filename})
            # NOTE: Filename in $DATA may be max 80 characters
            elif len(filename) > 80:
                raise ValueError(f'Filename for data record too long (>80 characters): {filename}')
            else:
                if "'" not in filename:
                    node = AttrTree.create('filename', {'QFILENAME': f"'{filename}'"})
                elif '"' not in filename:
                    node = AttrTree.create('filename', {'QFILENAME': f'"{filename}"'})
                else:
                    raise ValueError('Cannot have both " and \' in filename.')
            (pre, _, post) = self.root.partition('filename')
            root = AttrTree(self.root.rule, pre + (node,) + post)
        return self.replace(root=root)

    @property
    def ignore_character(self):
        """The comment character from ex IGNORE=C or None if not available."""
        char = None
        if self.root.first_branch('ignchar', 'char'):
            for option in self.root.subtrees('ignchar'):
                if char is not None:
                    raise ModelSyntaxError("Redefinition of ignore character in $DATA")
                char = str(option.subtree('char'))
                if len(char) == 3:  # It must be quoted
                    char = char[1:-1]

        return char

    def set_ignore_character(self, c: str):
        if c != self.ignore_character:
            char = c if len(c) == 1 else f'"{c}"'
            if len(c) == 1:
                char = c
            elif "'" not in c:
                char = f"'{c}'"
            elif '"' not in c:
                char = f'"{c}"'
            else:
                raise ValueError('Cannot have both " and \' in ignore character.')
            char_node = AttrTree.create('char', [{'CHAR': char}])
            node = AttrTree.create('ignchar', [{'IGNORE': 'IGNORE'}, {'EQUALS': '='}, char_node])
            root = self.root.remove('ignchar')
            newrec = self.replace(root=root).append_option_node(node)
            return newrec
        else:
            return self

    def set_ignore_character_from_header(self, label):
        """Set ignore character from a header label
        If s[0] is a-zA-Z set @
        else set s[0]
        """
        c = label[0]
        if c.isalpha():
            return self.set_ignore_character('@')
        else:
            return self.set_ignore_character(c)

    @property
    def null_value(self):
        """The value to replace for NULL (i.e. . etc) in the dataset
        note that only +,-,0 (meaning 0) and 1-9 are allowed
        """
        if subtree := self.root.first_branch('null', 'char'):
            char = str(subtree)
            if char == '+' or char == '-':
                return 0
            else:
                return float(char)
        else:
            return 0

    @property
    def ignore(self):
        filters = []
        for option in self.root.subtrees('ignore'):
            for filt in option.subtrees('filter'):
                filters.append(filt)
        return filters

    def remove_ignore(self):
        newroot = self.root.remove('ignore')
        return self.replace(root=newroot)

    @property
    def accept(self):
        filters = []
        for option in self.root.subtrees('accept'):
            for filt in option.subtrees('filter'):
                filters.append(filt)
        return filters

    def remove_accept(self):
        newroot = self.root.remove('accept')
        return self.replace(root=newroot)



from typing import Optional

from pharmpy.model import Model
from pharmpy.modeling import resample_data
from pharmpy.tools.bootstrap.results import calculate_results
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import ModelEntry, Task, Workflow, WorkflowBuilder
from pharmpy.workflows.results import ModelfitResults


def create_workflow(model: Model, results: Optional[ModelfitResults] = None, resamples: int = 1):
    """Run bootstrap tool

    Parameters
    ----------
    model : Model
        Pharmpy model
    results : ModelfitResults
        Results for model
    resamples : int
        Number of bootstrap resamples

    Returns
    -------
    BootstrapResults
        Bootstrap tool result object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model
    >>> from pharmpy.tools import run_bootstrap, load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> run_bootstrap(model, res, resamples=500) # doctest: +SKIP
    """

    wb = WorkflowBuilder(name='bootstrap')

    for i in range(resamples):
        task_resample = Task('resample', resample_model, model, f'bs_{i + 1}')
        wb.add_task(task_resample)

    wf_fit = create_fit_workflow(n=resamples)
    wb.insert_workflow(wf_fit)

    task_result = Task('results', post_process_results, results)
    wb.add_task(task_result, predecessors=wb.output_tasks)

    return Workflow(wb)


def resample_model(input_model, name):
    resample = resample_data(
        input_model, input_model.datainfo.id_column.name, resamples=1, name=name
    )
    model, _ = next(resample)
    model_entry = ModelEntry.create(model=model, parent=input_model)
    return model_entry


def post_process_results(original_model_res, *model_entries):
    models = [model_entry.model for model_entry in model_entries]
    modelfit_results = [model_entry.modelfit_results for model_entry in model_entries]
    res = calculate_results(
        models,
        results=modelfit_results,
        original_results=original_model_res,
        included_individuals=None,
        dofv_results=None,
    )
    return res



from __future__ import annotations

import math
from itertools import chain
from typing import TYPE_CHECKING, Iterable, Literal, Optional, Union

from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.expr.subs import subs, xreplace_dict
from pharmpy.internals.math import round_to_n_sigdig
from pharmpy.model import CompartmentalSystem, CompartmentalSystemBuilder, Model, output
from pharmpy.model.distributions.numeric import ConstantDistribution
from pharmpy.model.random_variables import (
    eval_expr,
    filter_distributions,
    sample_rvs,
    subs_distributions,
)

from .data import get_ids, get_observations
from .expressions import get_individual_parameters
from .odes import get_initial_conditions
from .parameter_sampling import create_rng, sample_parameters_from_covariance_matrix
from .parameters import get_omegas
from .random_variables import replace_non_random_rvs

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd
    import sympy
else:
    from pharmpy.deps import numpy as np
    from pharmpy.deps import pandas as pd
    from pharmpy.deps import sympy


def calculate_eta_shrinkage(
    model: Model,
    parameter_estimates: pd.Series,
    individual_estimates: pd.DataFrame,
    sd: bool = False,
):
    """Calculate eta shrinkage for each eta

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameter_estimates : pd.Series
        Parameter estimates
    individual_estimates : pd.DataFrame
        Table of individual (eta) estimates
    sd : bool
        Calculate shrinkage on the standard deviation scale (default is to calculate on the
        variance scale)

    Return
    ------
    Series
        Shrinkage for each eta

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> pe = results.parameter_estimates
    >>> ie = results.individual_estimates
    >>> calculate_eta_shrinkage(model, pe, ie)
    ETA_1    0.720481
    ETA_2    0.240295
    dtype: float64
    >>> calculate_eta_shrinkage(model, pe, ie, sd=True)
    ETA_1    0.471305
    ETA_2    0.128389
    dtype: float64

    See also
    --------
    calculate_individual_shrinkage

    """
    # Want parameter estimates combined with fixed parameter values
    param_inits = model.parameters.to_dataframe()['value']
    pe = parameter_estimates.combine_first(param_inits)

    param_names = model.random_variables.iiv.variance_parameters
    diag_ests = pe[param_names]
    diag_ests.index = individual_estimates.columns
    if not sd:
        shrinkage = 1 - (individual_estimates.var() / diag_ests)
    else:
        shrinkage = 1 - (individual_estimates.std() / (diag_ests**0.5))
    return shrinkage


def calculate_individual_shrinkage(
    model: Model, parameter_estimates: pd.Series, individual_estimates_covariance: pd.DataFrame
):
    """Calculate the individual eta-shrinkage

    Definition: ieta_shr = (var(eta) / omega)

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameter_estimates : pd.Series
        Parameter estimates of model
    individual_estimates_covariance : pd.DataFrame
        Uncertainty covariance matrices of individual estimates

    Return
    ------
    DataFrame
        Shrinkage for each eta and individual

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> pe = results.parameter_estimates
    >>> covs = results.individual_estimates_covariance
    >>> calculate_individual_shrinkage(model, pe, covs)
           ETA_1     ETA_2
    ID
    1   0.847789  0.256473
    2   0.796643  0.210669
    3   0.755025  0.226957
    4   0.764541  0.216405
    5   0.816192  0.203974
    6   0.778108  0.210992
    7   0.659420  0.236875
    8   0.668551  0.240097
    9   0.260056  0.200374
    10  0.725190  0.226563
    11  0.972110  0.421852
    12  0.249640  0.254119
    13  0.730294  0.364932
    14  0.165785  0.194464
    15  0.813399  0.313554
    16  0.797328  0.213211
    17  0.769059  0.278079
    18  0.098506  0.176778
    19  0.749022  0.235386
    20  0.742181  0.222932
    21  0.317956  0.264473
    22  0.943950  0.232732
    23  0.707183  0.259077
    24  0.553787  0.247717
    25  0.826349  0.114302
    26  0.854777  0.341384
    27  0.820829  0.263235
    28  0.999942  0.319986
    29  0.967084  0.432760
    30  0.404773  0.325215
    31  0.999980  0.318421
    32  0.925283  0.167667
    33  0.913706  0.242106
    34  0.875554  0.249197
    35  0.849135  0.294294
    36  0.172206  0.246422
    37  0.747380  0.278340
    38  0.187440  0.231249
    39  0.237805  0.254485
    40  0.999925  0.189793
    41  0.941906  0.170998
    42  0.923801  0.244046
    43  0.999928  0.320403
    44  0.237637  0.260453
    45  0.869540  0.194503
    46  0.999949  0.319750
    47  0.983782  0.393234
    48  0.698267  0.169337
    49  0.776674  0.214962
    50  0.688847  0.192608
    51  0.822213  0.202534
    52  0.511489  0.273601
    53  0.964757  0.223448
    54  0.762153  0.181648
    55  0.965657  0.435741
    56  0.995278  0.354798
    57  0.813382  0.263372
    58  0.727295  0.232867
    59  0.738777  0.224742

    See also
    --------
    calculate_eta_shrinkage

    """
    cov = individual_estimates_covariance
    pe = parameter_estimates
    # Want parameter estimates combined with fixed parameter values
    param_inits = model.parameters.to_dataframe()['value']
    pe = pe.combine_first(param_inits)

    # Get all iiv and iov variance parameters
    diag = model.random_variables.etas.covariance_matrix.diagonal()
    param_names = [s.name for s in diag]

    diag_ests = pe[param_names]

    def fn(row, ests):
        names = row[0].index
        ser = pd.Series(np.diag(row[0].values) / ests, index=names)
        return ser

    ish = pd.DataFrame(cov).apply(fn, axis=1, ests=diag_ests.values)
    return ish


def calculate_individual_parameter_statistics(
    model: Model,
    expr_or_exprs: Union[
        Iterable[sympy.Eq], Iterable[sympy.Expr], Iterable[str], sympy.Eq, sympy.Expr, str
    ],
    parameter_estimates: pd.Series,
    covariance_matrix: Optional[pd.DataFrame] = None,
    seed: Optional[Union[np.random.Generator, int]] = None,
):
    """Calculate statistics for individual parameters

    Calculate the mean (expected value of the distribution), variance
    (variance of the distribution) and standard error for individual
    parameters described by arbitrary expressions. Any dataset column or
    variable used in the model can be used in the expression. The exception
    being that variables that depends on the solution of the ODE system
    cannot be used. If covariates are used in the expression the statistics
    of the parameter is calculated at the median value of each covariate as well
    as at the 5:th and 95:th percentiles. If no parameter uncertainty is available
    for the model the standard error will not be calculated.

    Parameters
    ----------
    model : Model
        A previously estimated model
    parameter_estimates : pd.Series
        Parameter estimates
    covariance_matrix : pd.DataFrame
        Parameter uncertainty covariance matrix
    expr_or_exprs : str
        sympy expression or iterable of str or sympy expressions
        Expressions or equations for parameters of interest. If equations are used
        the names of the left hand sides will be used as the names of the parameters.
    seed : Generator or int
        Random number generator or int seed

    Returns
    -------
    pd.DataFrame
        A DataFrame of statistics indexed on parameter and covariate value.

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, create_rng
    >>> from pharmpy.modeling import calculate_individual_parameter_statistics
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> rng = create_rng(23)
    >>> pe = results.parameter_estimates
    >>> cov = results.covariance_matrix
    >>> calculate_individual_parameter_statistics(model, "K=CL/V", pe, cov, seed=rng)
                              mean  variance    stderr
    parameter covariates
    K         p5          0.004234  0.000001  0.001138
              median      0.004907  0.000001  0.001247
              p95         0.004907  0.000001  0.001247
    """
    rng = create_rng(seed)

    split_exprs = map(
        _split_equation,
        [expr_or_exprs]
        if isinstance(expr_or_exprs, str) or isinstance(expr_or_exprs, sympy.Basic)
        else expr_or_exprs,
    )

    full_exprs = list(
        map(
            lambda e: (e[0], model.statements.before_odes.full_expression(e[1])),
            split_exprs,
        )
    )

    input_parameter_estimates = parameter_estimates
    parameter_estimates = xreplace_dict(parameter_estimates)

    all_free_symbols = set().union(*map(lambda e: e[1].free_symbols, full_exprs))

    all_covariate_free_symbols = all_free_symbols.intersection(
        map(sympy.Symbol, model.datainfo.names)
    )
    all_parameter_free_symbols = set(parameter_estimates.keys())
    all_random_free_symbols = all_free_symbols.difference(
        all_parameter_free_symbols, all_covariate_free_symbols
    )

    distributions = list(
        filter_distributions(
            model.random_variables,
            all_random_free_symbols,
        )
    )

    sampling_rvs = list(
        subs_distributions(
            distributions,
            parameter_estimates,
        )
    )

    batches = []

    if not all_covariate_free_symbols:
        cases = {'median': {}}
    else:
        dataset = model.dataset
        column_filter = ['ID'] + list(symbol.name for symbol in all_covariate_free_symbols)
        q5 = dataset[column_filter].groupby('ID').median().quantile(0.05)
        q95 = dataset[column_filter].groupby('ID').median().quantile(0.95)
        median = dataset[column_filter].groupby('ID').median().median()
        cases = {
            'p5': xreplace_dict(q5),
            'median': xreplace_dict(median),
            'p95': xreplace_dict(q95),
        }

    filtered_sampling_rvs = list(
        filter(
            lambda r: any(map(all_random_free_symbols.__contains__, r[0])),
            sampling_rvs,
        )
    )

    nsamples = 1000000
    nbatches = 100
    batchsize = 10

    samples = sample_rvs(filtered_sampling_rvs, nsamples, rng)

    if covariance_matrix is not None:
        parameters_samples = sample_parameters_from_covariance_matrix(
            model,
            input_parameter_estimates,
            covariance_matrix,
            n=nbatches,
            force_posdef_covmatrix=True,
            seed=rng,
        )

        for _, row in parameters_samples.iterrows():
            parameters = xreplace_dict(row)
            local_sampling_rvs = list(subs_distributions(distributions, parameters)) + [
                ((key,), ConstantDistribution(value))
                for key, value in parameters.items()
                if key in all_parameter_free_symbols
            ]
            batch = sample_rvs(local_sampling_rvs, batchsize, rng)
            batches.append(batch)

    table = pd.DataFrame(columns=['parameter', 'covariates', 'mean', 'variance', 'stderr'])
    i = 0

    for name, full_expr in full_exprs:
        df = pd.DataFrame(index=list(cases.keys()), columns=['mean', 'variance', 'stderr'])
        parameter_estimates_expr = subs(full_expr, parameter_estimates, simultaneous=True)

        for case, cov_values in cases.items():
            expr = subs(parameter_estimates_expr, cov_values, simultaneous=True)
            values = eval_expr(expr, nsamples, samples)

            mean = np.mean(values)
            variance = np.var(values)

            # NOTE: This is NaN for empty inputs, dtype is required for those.
            cov_expr = subs(full_expr, cov_values, simultaneous=True)
            stderr = pd.Series(
                chain.from_iterable(eval_expr(cov_expr, batchsize, batch) for batch in batches),
                dtype='float64',
            ).std()

            df.loc[case] = [mean, variance, stderr]

        df.index.name = 'covariates'
        df.reset_index(inplace=True)
        if not name:
            name = f'unknown{i}'
            i += 1
        df['parameter'] = name
        table = pd.concat([table, df])

    table.set_index(['parameter', 'covariates'], inplace=True)
    table = table.astype('float64')

    return table


def calculate_pk_parameters_statistics(
    model: Model,
    parameter_estimates: pd.Series,
    covariance_matrix: Optional[pd.DataFrame] = None,
    seed: Optional[Union[np.random.Generator, int]] = None,
):
    """Calculate statistics for common pharmacokinetic parameters

    Calculate the mean (expected value of the distribution), variance
    (variance of the distribution) and standard error for some individual
    pre-defined pharmacokinetic parameters.

    Parameters
    ----------
    model : Model
        A previously estimated model
    parameter_estimates : pd.Series
        Parameter estimates
    covariance_matrix : pd.DataFrame
        Parameter uncertainty covariance matrix
    seed : Generator or int
        Random number generator or seed

    Returns
    -------
    pd.DataFrame
        A DataFrame of statistics indexed on parameter and covariate value.

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, create_rng
    >>> from pharmpy.modeling import calculate_pk_parameters_statistics
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> rng = create_rng(23)
    >>> pe = results.parameter_estimates
    >>> cov = results.covariance_matrix
    >>> calculate_pk_parameters_statistics(model, pe, cov, seed=rng)
                                  mean     variance     stderr
    parameter   covariates
    t_half_elim p5          173.337164  1769.493756  42.843398
                median      149.567842  1317.474199  36.233070
                p95         149.567842  1317.474199  36.233070
    k_e         p5            0.004234     0.000001   0.001138
                median        0.004907     0.000001   0.001247
                p95           0.004907     0.000001   0.001247


    See Also
    --------
    calculate_individual_parameter_statistics : Calculation of statistics for arbitrary parameters

    """

    statements = model.statements
    odes = statements.ode_system
    central = odes.central_compartment
    depot = odes.find_depot(statements)
    peripherals = odes.find_peripheral_compartments()
    elimination_rate = odes.get_flow(central, output)

    expressions = []  # Eq(name, expr)
    # FO abs + 1comp + FO elimination
    if len(odes) == 2 and depot and odes.t not in elimination_rate.free_symbols:
        ode_list, ics = odes.eqs, get_initial_conditions(model, dosing=True)
        sols = sympy.dsolve(ode_list, ics=ics)
        expr = sols[1].rhs
        d = sympy.diff(expr, odes.t)
        tmax_closed_form = sympy.solve(d, odes.t)[0]
        expressions.append(sympy.Eq(sympy.Symbol('t_max'), tmax_closed_form))
        e2 = sympy.simplify(expr / depot.doses[0].amount / sympy.denom(elimination_rate))
        cmax_dose_closed_form = sympy.simplify(
            subs(e2, {odes.t: tmax_closed_form}, simultaneous=True)
        )
        expressions.append(sympy.Eq(sympy.Symbol('C_max_dose'), cmax_dose_closed_form))

    # Any abs + 1comp + FO elimination
    if not peripherals and odes.t not in elimination_rate.free_symbols:
        elimination_system = statements.ode_system
        for name in elimination_system.compartment_names:
            if name != central.name:  # NOTE: Keep central
                cb = CompartmentalSystemBuilder(elimination_system)
                cb.remove_compartment(elimination_system.find_compartment(name))
                elimination_system = CompartmentalSystem(cb)
        eq = elimination_system.eqs[0]
        ic = sympy.Function(elimination_system.amounts[0].name)(0)
        A0 = sympy.Symbol('A0')
        sols = sympy.dsolve(eq, ics={ic: A0})
        eq_half = sympy.Eq(sympy.Rational(1, 2) * A0, sols.rhs)
        thalf_elim = sympy.solve(eq_half, odes.t)[0]
        expressions.append(sympy.Eq(sympy.Symbol('t_half_elim'), thalf_elim))

    # Bolus dose + 2comp + FO elimination
    if len(peripherals) == 1 and len(odes) == 2 and odes.t not in elimination_rate.free_symbols:
        ode_list, ics = odes.eqs, get_initial_conditions(model, dosing=True)
        sols = sympy.dsolve(ode_list, ics=ics)
        A = sympy.Wild('A')
        B = sympy.Wild('B')
        alpha = sympy.Wild('alpha')
        beta = sympy.Wild('beta')

        m = sols[0].rhs.match(A * sympy.exp(-alpha) + B * sympy.exp(-beta))

        beta = m[beta] / odes.t
        alpha = m[alpha] / odes.t
        A = m[A] / central.doses[0].amount
        B = m[B] / central.doses[0].amount

        if (beta - alpha).extract_multiplicatively(-1) is not None:
            # alpha > beta  (sympy couldn't simplify this directly)
            alpha, beta = beta, alpha
            A, B = B, A
        expressions.append(sympy.Eq(sympy.Symbol('A'), A))
        expressions.append(sympy.Eq(sympy.Symbol('B'), B))
        expressions.append(sympy.Eq(sympy.Symbol('alpha'), alpha))
        expressions.append(sympy.Eq(sympy.Symbol('beta'), beta))

    # Any abs + any comp + FO elimination
    if odes.t not in elimination_rate.free_symbols:
        expressions.append(sympy.Eq(sympy.Symbol('k_e'), elimination_rate))

    df = calculate_individual_parameter_statistics(
        model, expressions, parameter_estimates, covariance_matrix, seed=seed
    )
    return df


def _split_equation(s):
    if isinstance(s, str):
        a = s.split('=')
        if len(a) == 1:
            name = None
            expr = parse_expr(s)
        else:
            name = a[0].strip()
            expr = parse_expr(a[1])
    elif isinstance(s, sympy.Eq):
        name = s.lhs.name
        expr = s.rhs
    else:  # sympy expr
        name = None
        expr = s
    if name is None and isinstance(expr, sympy.Symbol):
        name = expr.name
    return name, expr


def calculate_aic(model: Model, likelihood: float):
    """Calculate AIC

    AIC = -2LL + 2*n_estimated_parameters

    Parameters
    ----------
    model : Model
        Pharmpy model object
    likelihood : float
        -2LL

    Returns
    -------
    float
        AIC of model fit
    """
    parameters = model.parameters.nonfixed
    return likelihood + 2 * len(parameters)


def _random_etas(model):
    var = model.random_variables.etas.variance_parameters
    zerofix = [model.parameters[e].fix and model.parameters[e].init == 0 for e in var]
    keep = []
    for eta, zf in zip(model.random_variables.etas.names, zerofix):
        if not zf:
            keep.append(eta)
    return model.random_variables.etas[keep]


def calculate_bic(
    model: Model,
    likelihood: float,
    type: Literal['mixed', 'fixed', 'random', 'iiv'] = 'mixed',
    multiple_testing: bool = False,
    mult_test_p: int = 1,
    mult_test_e: int = 1,
):
    """Calculate BIC

    Different variations of the BIC can be calculated:

    * | mixed (default)
      | BIC = -2LL + n_random_parameters * log(n_individuals) +
      |       n_fixed_parameters * log(n_observations)
    * | fixed
      | BIC = -2LL + n_estimated_parameters * log(n_observations)
    * | random
      | BIC = -2LL + n_estimated_parameters * log(n_individals)
    * | iiv
      | BIC = -2LL + n_estimated_iiv_omega_parameters * log(n_individals)

    If multiple_testing option is set to true an additional penalty will be added:

    * | mBIC = BIC + 2*(n_estimated_parameters)*log(n_predictors/n_expected_models)

    Parameters
    ----------
    model : Model
        Pharmpy model object
    likelihood : float
        -2LL to use
    type : {'mixed', 'fixed', 'random', 'iiv'}
        Type of BIC to calculate. Default is the mixed effects.
    multiple_testing : bool
        Whether to use penalty for multiple testing (default is False)
    mult_test_p : int
        Number of expected models if using type `mult_test`
    mult_test_e : int
        E value if using type `mult_test`

    Returns
    -------
    float
        BIC of model fit

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> ofv = results.ofv
    >>> calculate_bic(model, ofv)
    611.7071686183284
    >>> calculate_bic(model, ofv, type='fixed')
    616.536606983396
    >>> calculate_bic(model, ofv, type='random')
    610.7412809453149
    >>> calculate_bic(model, ofv, type='iiv')
    594.431131169692
    """
    parameters = model.parameters.nonfixed
    if type == 'fixed':
        penalty = len(parameters) * math.log(len(get_observations(model)))
    elif type == 'random':
        penalty = len(parameters) * math.log(len(get_ids(model)))
    elif type == 'iiv':
        nomegas_iiv = len(
            [name for name in model.random_variables.iiv.parameter_names if name in parameters]
        )
        penalty = nomegas_iiv * math.log(len(get_ids(model)))
    elif type == 'mixed':
        theta_f, theta_r = _categorize_parameters(model)
        nsubs = len(get_ids(model))
        nobs = len(get_observations(model))
        penalty = len(theta_r) * math.log(nsubs) + len(theta_f) * math.log(nobs)
    else:
        supported_types = ('mixed', 'fixed', 'random', 'iiv')
        raise ValueError(f'Unknown `type` \'{type}\', must be one of: {supported_types}')

    if multiple_testing:
        if mult_test_p <= 0 or mult_test_e <= 0:
            raise ValueError(
                'Options `mult_test_p` and `mult_test_e` must be >= 0 for method `mult_test`'
            )
        penalty += 2 * len(parameters) * math.log(mult_test_p / mult_test_e)

    return likelihood + penalty


def _categorize_parameters(model):
    # Categorize parameters into random and fixed
    model = replace_non_random_rvs(model)  # To get rid of 0 FIX rvs
    indpars = get_individual_parameters(model)
    all_pop_params = set(model.parameters.nonfixed.symbols)
    omegas = set(get_omegas(model).symbols)
    etas = set(model.random_variables.etas.symbols)
    epsilons = set(model.random_variables.epsilons.symbols)
    randpars = omegas.intersection(all_pop_params)
    fixedpars = set()
    for indpar in indpars:
        expr = model.statements.before_odes.full_expression(indpar)
        symbols = expr.free_symbols
        param_symbols = symbols.intersection(all_pop_params)
        if symbols.isdisjoint(etas):
            fixedpars |= param_symbols - randpars
        else:
            fixedpars -= param_symbols
            randpars |= param_symbols

    for y in model.dependent_variables.keys():
        expr = model.statements.after_odes.full_expression(y)
        symbols = expr.free_symbols
        param_symbols = symbols.intersection(all_pop_params)
        cureps = symbols.intersection(epsilons)
        cursigmas = model.random_variables[cureps].free_symbols - cureps
        cursigmas = cursigmas.intersection(all_pop_params)
        cursymbols = param_symbols | cursigmas
        if symbols.isdisjoint(etas):
            fixedpars |= cursymbols - randpars
        else:
            fixedpars -= cursymbols
            randpars |= cursymbols
    return fixedpars, randpars


def check_high_correlations(model: Model, cor: pd.DataFrame, limit: float = 0.9):
    """Check for highly correlated parameter estimates

    Parameters
    ----------
    model : Model
        Pharmpy model object
    cor : pd.DataFrame
        Estimated correlation matrix
    limit : float
        Lower limit for a high correlation

    Returns
    -------
    pd.Series
        Correlation values indexed on pairs of parameters for (absolute) correlations above limit

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> cor = results.correlation_matrix
    >>> check_high_correlations(model, cor, limit=0.3)
    PTVCL  IVCL      -0.388059
    PTVV   THETA_3   -0.356899
           IVV        0.356662
    dtype: float64
    """
    high_and_below_diagonal = cor.abs().ge(limit) & np.triu(np.ones(cor.shape), k=1).astype(bool)
    return cor.where(high_and_below_diagonal).stack()


def check_parameters_near_bounds(
    model: Model, values: pd.Series, zero_limit: float = 0.001, significant_digits: int = 2
):
    """Check if any estimated parameter value is close to its bounds

    Parameters
    ----------
    model : Model
        Pharmpy model object
    values : pd.Series
        Series of values with index a subset of parameter names.
    zero_limit : float
        maximum distance to 0 bounds
    significant_digits : int
        maximum distance to non-zero bounds in number of significant digits

    Returns
    -------
    pd.Series
        Logical Series with same index as values

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> from pharmpy.tools import load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> check_parameters_near_bounds(model, results.parameter_estimates)
    PTVCL        False
    PTVV         False
    THETA_3      False
    IVCL         False
    IVV          False
    SIGMA_1_1    False
    dtype: bool

    """
    ser = pd.Series(
        [
            _is_close_to_bound(model.parameters[p], values.loc[p], zero_limit, significant_digits)
            for p in values.index
        ],
        index=values.index,
        dtype=bool,
    )
    return ser


def _is_close_to_bound(param, value=None, zero_limit=0.01, significant_digits=2):
    if value is None:
        value = param.init
    return (
        param.lower > -sympy.oo
        and _is_near_target(value, param.lower, zero_limit, significant_digits)
    ) or (
        param.upper < sympy.oo
        and _is_near_target(value, param.upper, zero_limit, significant_digits)
    )


def _is_near_target(x, target, zero_limit, significant_digits):
    if target == 0:
        return abs(x) < abs(zero_limit)
    else:
        return round_to_n_sigdig(x, n=significant_digits) == round_to_n_sigdig(
            target, n=significant_digits
        )



import shutil

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Model
from pharmpy.tools import read_modelfit_results, run_tool


def test_bootstrap(tmp_path, testdata):
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.ext', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.lst', tmp_path)
        model = Model.parse_model('pheno.mod')
        results = read_modelfit_results('pheno.mod')
        model = model.replace(datainfo=model.datainfo.replace(path=tmp_path / 'pheno.dta'))
        res = run_tool('bootstrap', model=model, results=results, resamples=3)
        assert len(res.parameter_estimates) == 3
